import numpy as np
from UncertainSCI.opoly1d import eval_driver


def verify_orthonormal(ab, n, xg, wg):
    nmax = np.max(n)
    assert nmax <= len(ab) - 1
    P = eval_driver(xg, n, 0, ab)
    return np.dot(wg*P.T, P)



"""
Contains routines that specialize opoly1d things for classical orthogonal polynomial families
- jacobi polys
- hermite poly
- laguerre polys
"""
import numpy as np

from UncertainSCI.opoly1d import OrthogonalPolynomialBasis1D
from UncertainSCI.opoly1d import eval_driver, idistinv_driver, gauss_quadrature_driver
from UncertainSCI.opoly1d import linear_modification, quadratic_modification
from UncertainSCI.transformations import AffineTransform
from UncertainSCI.utils.casting import to_numpy_array

import os
import pickle
import UncertainSCI as uSCI

from scipy import special as sp
from scipy.stats import beta as bbeta
from scipy.stats import gamma

def jacobi_weight_normalized(x, alpha, beta):
    """
    Evaluates the Jacobi weight function defined as 

      w(x) = C(alpha,beta) * (1-x)**alpha, (1+x)**beta,

      1/C(alpha,beta) = B(beta+1,alpha+1) * 2**(alpha+beta+1)

    for alpha, beta > -1, and |x| <= 1. This weight function is a probability
    density on [-1,1].
    """

    return 1/(2**(alpha+beta+1)*sp.beta(beta+1,alpha+1)) * (1-x)**alpha * (1+x)**beta

def jacobi_recurrence_values(N, alpha, beta):
    """
    Returns the first N+1 recurrence coefficient pairs for the (alpha, beta)
    Jacobi family
    """
    if N < 1:
        ab = np.ones((1, 2))
        ab[0, 0] = 0
        ab[0, 1] = np.exp((alpha + beta + 1.) * np.log(2.) +
                          sp.gammaln(alpha + 1.) + sp.gammaln(beta + 1.) -
                          sp.gammaln(alpha + beta + 2.))
        ab = np.sqrt(ab)
        return ab

    ab = np.ones((N+1, 2)) * np.array([beta**2. - alpha**2., 1.])

    # Special cases
    ab[0, 0] = 0.
    ab[1, 0] = (beta - alpha) / (alpha + beta + 2.)
    ab[0, 1] = np.exp((alpha + beta + 1.) * np.log(2.) +
                      sp.gammaln(alpha + 1.) + sp.gammaln(beta + 1.) -
                      sp.gammaln(alpha + beta + 2.))

    ab[1, 1] = 4. * (alpha + 1.) * (beta + 1.) / (
                   (alpha + beta + 2.)**2 * (alpha + beta + 3.))

    if N > 1:
        ab[1, 1] = 4. * (alpha + 1.) * (beta + 1.) / (
                   (alpha + beta + 2.)**2 * (alpha + beta + 3.))

        ab[2, 0] /= (2. + alpha + beta) * (4. + alpha + beta)
        inds = 2
        ab[2, 1] = 4 * inds * (inds + alpha) * (inds + beta) * (inds + alpha + beta)
        ab[2, 1] /= (2. * inds + alpha + beta)**2 * (2. * inds + alpha + beta + 1.) * (2. * inds + alpha + beta - 1)

    if N > 2:
        inds = np.arange(2., N+1)
        ab[3:, 0] /= (2. * inds[:-1] + alpha + beta) * (2 * inds[:-1] + alpha + beta + 2.)
        ab[2:, 1] = 4 * inds * (inds + alpha) * (inds + beta) * (inds + alpha + beta)
        ab[2:, 1] /= (2. * inds + alpha + beta)**2 * (2. * inds + alpha + beta + 1.) * (2. * inds + alpha + beta - 1)

    ab[:, 1] = np.sqrt(ab[:, 1])

    return ab


def jacobi_idist_driver(x, n, alpha, beta, M):

    A = int(np.floor(np.abs(alpha)))
    Aa = alpha - A

    if isinstance(x, float) or isinstance(x, int):
        x = np.asarray([x])
    else:
        x = np.asarray(x)

    F = np.zeros(x.size)

    ab = jacobi_recurrence_values(n, alpha, beta)
    ab[0, 1] = 1

    if n > 0:
        xn, wn = gauss_quadrature_driver(ab, n)

    """
    This is the (inverse) n'th root of the leading coefficient square of p_n
    """
    if n == 0:
        kn_factor = 0  # could be any value since we don't use it when n = 0
    else:
        kn_factor = np.exp(-1/n * np.sum(np.log(ab[:, 1]**2)))

    for ind in range(x.size):
        if x[ind] == -1:
            F[ind] = 0
            continue

        ab = jacobi_recurrence_values(n+A+M, 0, beta)
        ab[0, 1] = 1.

        if n > 0:
            un = (2./(x[ind]+1.)) * (xn + 1.) - 1.

        logfactor = 0.
        for j in range(n):
            # ab = quad_mod(ab, un[j])
            ab = quadratic_modification(ab, un[j])
            logfactor = logfactor + np.log(ab[0, 1]**2 * ((x[ind]+1)/2)**2 * kn_factor)
            ab[0, 1] = 1.

        root = (3.-x[ind]) / (1.+x[ind])

        for k in range(A):
            # ab = lin_mod(ab, root)
            ab = linear_modification(ab, root)
            logfactor = logfactor + np.log(ab[0, 1]**2 * 1/2 * (x[ind]+1))
            ab[0, 1] = 1.

        u, w = gauss_quadrature_driver(ab, M)

        Ival = np.sum(w * (2 - 1/2 * (u+1) * (x[ind]+1))**Aa)
        F[ind] = np.exp(logfactor - alpha*np.log(2) - sp.betaln(beta+1, alpha+1) -
                        np.log(beta+1) + (beta+1) * np.log((x[ind]+1)/2)) * Ival

    return F


def fidistinv_setup_helper1(ug, exps):

    if isinstance(ug, float) or isinstance(ug, int):
        ug = np.asarray([ug])
    else:
        ug = np.asarray(ug)

    ug_mid = 1/2 * (ug[:-1] + ug[1:])
    ug = np.sort(np.append(ug, ug_mid))

    exponents = np.zeros((2, ug.size-1))

    exponents[0, ::2] = 2/3
    exponents[1, 1::2] = 2/3

    exponents[0, 0] = exps[0]
    exponents[-1, -1] = exps[1]

    return ug, exponents


def fidistinv_setup_helper2(ug, idistinv, exponents, M, alpha, beta):

    # vgrid = np.cos( np.linspace(np.pi, 0, M) )
    # Or
    # vgrid = np.cos( np.linspace(np.pi, 0, M+1)[:-1] + M/(2*np.pi) )
    # Or
    vgrid = np.cos(np.linspace(np.pi, 0, M+2))
    vgrid = vgrid[1:-1]

    ab = jacobi_recurrence_values(M, -1/2, -1/2)
    V = eval_driver(vgrid, np.arange(M), 0, ab)

    iV = np.linalg.inv(V)

    ugrid = np.zeros((M, ug.size - 1))
    xgrid = np.zeros((M, ug.size - 1))
    xcoeffs = np.zeros((M, ug.size - 1))

    for q in range(ug.size - 1):
        ugrid[:, q] = (vgrid + 1) / 2 * (ug[q+1] - ug[q]) + ug[q]

        if ug.size == 3:
            xgrid[:, q] = 2 * bbeta.ppf(ugrid[:, q], beta+1, alpha+1) - 1
        else:
            xgrid[:, q] = idistinv(ugrid[:, q])

        temp = xgrid[:, q]

        # temp = temp(ugrid)
        # for ugrid near 0, then temp behaves like a certain rational function (1-ugrid)**???
        if exponents[0, q] != 0:
            temp = (temp - xgrid[0, q]) / (xgrid[-1, q] - xgrid[0, q])
        else:
            temp = (temp - xgrid[-1, q]) / (xgrid[-1, q] - xgrid[0, q])

        with np.errstate(divide='ignore', invalid='ignore'):
            temp = temp * (1 + vgrid)**exponents[0, q] * (1 - vgrid)**exponents[1, q]
            temp[~np.isfinite(temp)] = 0

        xcoeffs[:, q] = np.dot(iV, temp)

    data = np.zeros((M+6, ug.size - 1))
    for q in range(ug.size - 1):
        data[:, q] = np.hstack((ug[q], ug[q+1], xgrid[0, q], xgrid[-1, q], exponents[:, q], xcoeffs[:, q]))

    return data


def fidistinv_driver(u, n, data):

    if isinstance(u, float) or isinstance(u, int):
        u = np.asarray([u])
    else:
        u = np.asarray(u)

    if isinstance(n, float) or isinstance(n, int):
        n = np.asarray([n])
    else:
        n = np.asarray(n)

    if u.size == 0:
        return np.zeros(0)

    if n.size != 1:
        assert u.size == n.size  # Inputs u and n must be the same size, or n must be a scalar

    N = max(n)
    assert len(data) >= N+1  # Input data does not cover range of n

    x = np.zeros(u.size)
    if n.size == 1:
        x = driver_helper(u, data[int(n)])
    else:
        for q in range(N+1):
            nmask = (n == q)
            x[nmask] = driver_helper(u[nmask], data[q])

    return x


def driver_helper(u, data):

    tol = 1e-12

    M = data.shape[0] - 6

    ab = jacobi_recurrence_values(M, -1/2, -1/2)

    app = np.append(data[0, :], data[1, -1])
    edges = np.insert(app, [0, app.size], [-np.inf, np.inf])
    j = np.digitize(u, edges, right=False)
    B = edges.size - 1

    x = np.zeros(u.size)
    x[np.where(j == 1)] = data[2, 0]  # Boundary bins
    x[np.where(j == B)] = data[3, -1]

    for qb in range(2, B):
        umask = (j == qb)
        if not any(umask):
            continue

        q = qb - 1
        vgrid = (u[umask] - data[0, q-1]) / (data[1, q-1] - data[0, q-1]) * 2 - 1
        V = eval_driver(vgrid, np.arange(M), 0, ab)
        temp = np.dot(V, data[6:, q-1])
        with np.errstate(divide='ignore', invalid='ignore'):
            temp = temp / ((1 + vgrid)**data[4, q-1] * (1 - vgrid)**data[5, q-1])

        if data[4, q-1] != 0:
            flags = abs(u[umask] - data[0, q-1]) < tol
            temp[flags] = 0
            temp = temp * (data[3, q-1] - data[2, q-1]) + data[2, q-1]
        else:
            flags = abs(u[umask] - data[1, q-1]) < tol
            temp[flags] = 0
            temp = temp * (data[3, q-1] - data[2, q-1]) + data[3, q-1]

        x[umask] = temp

    return x


class JacobiPolynomials(OrthogonalPolynomialBasis1D):
    """
    Jacobi Polynomial family.
    """
    def __init__(self, alpha=0., beta=0., domain=[-1., 1.], **options):
        OrthogonalPolynomialBasis1D.__init__(self, **options)
        assert alpha > -1., beta > -1.
        self.alpha, self.beta = alpha, beta

        assert len(domain) == 2
        self.domain = np.array(domain).reshape([2, 1])
        self.standard_domain = np.array([-1, 1]).reshape([2, 1])
        self.transform_to_standard = AffineTransform(domain=self.domain, image=self.standard_domain)

    def recurrence_driver(self, N):
        # Returns the first N+1 recurrence coefficient pairs for the Jacobi
        # polynomial family.
        ab = jacobi_recurrence_values(N, self.alpha, self.beta)
        if self.probability_measure and N >= 0:
            ab[0, 1] = 1.

        return ab

    def idist_medapprox(self, n):
        """
        Computes an approximation to the median of the degree-n induced
        distribution.
        """

        if n > 0:
            m = (self.beta**2-self.alpha**2) / (2*n+self.alpha+self.beta)**2
        else:
            m = 2/(1 + (self.alpha+1) / (self.beta+1)) - 1
        return m

    def idist(self, x, n, M=50):
        """
        Computes the order-n induced distribution at the locations x using M=10
        quadrature nodes.
        """
        if isinstance(x, float) or isinstance(x, int):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        F = np.zeros(x.size, )
        mrs_centroid = self.idist_medapprox(n)
        F[np.where(x <= mrs_centroid)] = jacobi_idist_driver(x[np.where(x <= mrs_centroid)], n, self.alpha, self.beta, M)
        F[np.where(x > mrs_centroid)] = 1 - jacobi_idist_driver(-x[np.where(x > mrs_centroid)], n, self.beta, self.alpha, M)

        return F

    def idistinv(self, u, n):

        if isinstance(u, float) or isinstance(u, int):
            u = np.asarray([u])
        else:
            u = np.asarray(u)

        x = np.zeros(u.size)
        supp = [-1, 1]

        if isinstance(n, float) or isinstance(n, int):
            n = np.asarray([n])
        else:
            n = np.asarray(n)

        M = 10

        if n.size == 1:
            n = int(n)

            def primitive(xx):
                return self.idist(xx, n, M=M)

            ab = self.recurrence_driver(2*n + M+1)
            x = idistinv_driver(u, n, primitive, ab, supp)

        else:

            nmax = np.amax(n)
            ind = np.digitize(n, np.arange(-0.5, 0.5+nmax+1e-8), right=False)

            ab = self.recurrence_driver(2*nmax + M+1)
            for i in range(nmax+1):
                flags = ind == i+1

                def primitive(xx):
                    return self.idist(xx, i, M=M)

                x[flags] = idistinv_driver(u[flags], i, primitive, ab, supp)

        return x

    def fidistinv_jacobi_setup(self, n, data):
        ns = np.arange(len(data), n+1)

        for q in range(ns.size):

            nn = ns[q]

            if nn == 0:
                ug = np.array([0, 1])
            else:
                xg, wg = self.gauss_quadrature(nn)
                ug = self.idist(xg, nn)
                ug = np.insert(ug, [0, ug.size], [0, 1])

            # F_n^{-1}(u) near u = 0 (also u = 1)
            # F_n(x) = \int_{-1}^x p_n^2(s) w(s) ds
            # u near 0 =====> x near -1
            # near -1: F_n(x) = D(alpha,beta) \int_{-1}^{-1+eps} p_n^2(s) w(s) ds
            #
            #        D(alpha,beta)^{-1} = 2^(alpha+beta+1) B(beta+1, alpha+1)
            #
            #          p_n^2(s) \sim C_n = p_n^2(-1) near s = -1 (C_n is computable)
            #
            # near -1: F_n(x) \approx D \int_{-1}^{-1+eps} C_n (1-s)^alpha (1+s)^beta ds
            #                 \approx D \int_{-1}^{-1+eps} C_n 2^alpha (1+s)^beta ds
            #                 = 2^alpha D C_n \int_{-1}^{-1+eps} (1+s)^beta ds
            #                 = 2^alpha D C_n \int_{0}^{eps} r^beta dr
            #                 = 2^alpha D C_n / (beta+1) eps^(beta+1)  (beta + 1 > 0)
            #
            # near -1: F_n(x) \approx 2^alpha D C_n  (1+x)^(beta+1)
            # near u = 0: u \approx 2^alpha D C_n  (1+x)^(beta+1)
            # near u = 0: (u / (2^alpha * D C_n))^(1/(beta+1)) - 1 = x
            #
            # idistinv_helper: encode computation x = (u / (2^alpha * D * C_n))^(1/(beta+1)) - 1 (near x = -1)
            #                                     x = u^(1/(beta+1)) / E_n   - 1
            #                                           E_n = (1 / (2^alpha * D * C_n))^(1/(beta+1))
            #                                      E_n (probably) should be computed with logs:
            #                                       E_n = exp( -1/(beta+1) * (alpha*log(2) + log(D) + log(C_n) ) )
            #                                           For D: there is a logbeta function
            # temp(vgrid) ~ (1 + ugrid)^(beta + 1)
            exps = np.array([1/(self.beta+1), 1/(self.alpha+1)])
            ug, exponents = fidistinv_setup_helper1(ug, exps)

            def idistinv(u):
                return self.idistinv(u, nn)

            data.append(fidistinv_setup_helper2(ug, idistinv, exponents, 10, self.alpha, self.beta))  # , E_n?

        return data

    def fidistinv(self, u, n):

        dirName = 'data_set'
        parent_dir = os.path.dirname(os.path.dirname(uSCI.__file__))
        path = os.path.join(parent_dir, dirName)
        # Path(path).mkdir(parents=True, exist_ok=True)

        try:
            os.makedirs(path)
            print('Directory', dirName, 'created')
        except FileExistsError:
            pass
            # print ('Directory ', dirName, 'already exists')

        filename = 'data_jacobi_{0:1.6f}_{1:1.6f}'.format(self.alpha, self.beta)
        try:
            with open(os.path.join(path, filename), 'rb') as f:
                data = pickle.load(f)
        except Exception:
            data = []
            with open(os.path.join(path, filename), 'ab+') as f:
                pickle.dump(data, f)

        if isinstance(n, float) or isinstance(n, int):
            n = np.asarray([n])
        else:
            n = np.asarray(n)

        if len(data) < max(n[:]) + 1:
            msg = 'Precomputing data for Jacobi parameters (alpha,beta) = ({0:1.6f}, {1:1.6f})...'
            print(msg.format(self.alpha, self.beta), end='', flush=True)
            data = self.fidistinv_jacobi_setup(max(n[:]), data)
            with open(os.path.join(path, filename), 'wb') as f:
                pickle.dump(data, f)
            print('Done', flush=True)

        x = fidistinv_driver(u, n, data)

        return x

    def eval_1d(self, x, n):
        """
        Evaluates univariate orthonormal polynomials given their
        three-term recurrence coefficients ab

        """

        n = np.asarray(n)
        if isinstance(x, int) or isinstance(x, float):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        if n.size < 1 or x.size < 1:
            return np.zeros(0)

        nmax = np.max(n)

        ab = self.recurrence_driver(nmax+1)

        assert ab.shape[0] > nmax
        assert np.min(n) > -1

        p = np.zeros(x.shape + (nmax+1,))
        xf = x.flatten()

        p[:, 0] = 1/ab[0, 1]

        if nmax > 0:
            p[:, 1] = 1/ab[1, 1] * ((xf - ab[1, 0])*p[:, 0])

        for j in range(2, nmax+1):
            p[:, j] = 1/ab[j, 1] * ((xf - ab[j, 0])*p[:, j-1] - ab[j-1, 1]*p[:, j-2])

        return p[:, n.flatten()]

    def eval_nd(self, x, lambdas):
        """
        Evaluates tensorial orthonormal polynomials associated with the
        univariate recurrence coefficients ab
        """

        try:
            M, d = x.shape
        except Exception:
            d = x.size
            M = 1
            x = np.reshape(x, (M, d))

        N, d2 = lambdas.shape

        assert d == d2, "Dimension 1 of x and lambdas must be equal"

        p = np.ones([M, N])

        for qd in range(d):
            p = p * self.eval_1d(x[:, qd], lambdas[:, qd])

        return p


def hermite_recurrence_values(N, mu):
    ab = np.zeros((N+1, 2))
    ab[0, 1] = sp.gamma(mu + 1/2)
    ab[1:, 1] = 1/2 * np.arange(1, N+1)
    ab[np.arange(N+1) % 2 == 1, 1] += mu
    ab[:, 1] = np.sqrt(ab[:, 1])

    return ab


def freud_idist(x, n, alpha, rho):
    """
    for x <= 0
    """

    if isinstance(x, float) or isinstance(x, int):
        x = np.asarray([x])
    else:
        x = np.asarray(x)

    F = np.zeros(x.size)

    if n % 2 == 0:
        F = 1/2 * hfreud_idistc(x**2, int(n/2), alpha/2, (rho-1)/2)
    else:
        F = 1/2 * hfreud_idistc(x**2, int((n-1)/2), alpha/2, (rho+1)/2)

    return F


def freud_idistinv(u, n, alpha, rho):

    if isinstance(u, float) or isinstance(u, int):
        u = np.asarray([u])
    else:
        u = np.asarray(u)

    if isinstance(n, float) or isinstance(n, int):
        n = np.asarray([n])
    else:
        n = np.asarray(n)

    x = np.zeros(u.shape)

    if n.size == 1:
        n = n[0]
        if n % 2 == 0:
            x = np.sqrt(hfreud_idistinv(np.abs(2*u-1), int(n/2), alpha/2, (rho-1)/2))
        else:
            x = np.sqrt(hfreud_idistinv(np.abs(2*u-1), int((n-1)/2), alpha/2, (rho+1)/2))
    else:
        assert n.size == u.size
        evenflags = np.where(n % 2 == 0)
        oddflags = np.where(n % 2 != 0)
        x[evenflags] = np.sqrt(hfreud_idistinv(np.abs(2*u[evenflags]-1), (n[evenflags]/2).astype(int), alpha/2, (rho-1)/2))
        x[oddflags] = np.sqrt(hfreud_idistinv(np.abs(2*u[oddflags]-1), ((n[oddflags]-1)/2).astype(int), alpha/2, (rho+1)/2))

#     if n % 2 == 0:
#         x = np.sqrt( hfreud_idistinv(np.abs(2*u-1), int(n/2), alpha/2, (rho-1)/2) )
#     else:
#         x = np.sqrt( hfreud_idistinv(np.abs(2*u-1), int((n-1)/2), alpha/2, (rho+1)/2) )

    ind = np.where(u < 0.5)
    x[ind] = -x[ind]

    return x


class HermitePolynomials(OrthogonalPolynomialBasis1D):
    def __init__(self, alpha=2, rho=0., **options):
        OrthogonalPolynomialBasis1D.__init__(self, **options)
        assert rho > -1.
        self.alpha = alpha
        self.rho = rho

    def recurrence_driver(self, N):

        ab = hermite_recurrence_values(N, self.rho/2)
        if self.probability_measure and N > 0:
            ab[0, 1] = 1.

        return ab

    def idist(self, x, n):

        alpha = self.alpha
        rho = self.rho

        if isinstance(x, float) or isinstance(x, int):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        F = np.zeros(x.size)
        F[np.where(x <= 0)] = freud_idist(x[np.where(x <= 0)], n, alpha, rho)
        F[np.where(x > 0)] = 1 - freud_idist(-x[np.where(x > 0)], n, alpha, rho)

        return F

    def idistinv(self, u, n):

        alpha = self.alpha
        rho = self.rho

        return freud_idistinv(u, n, alpha, rho)


def laguerre_recurrence_values(N, alpha, rho):
    # Returns the first N+1 recurrence coefficient pairs for the Laguerre family.
    assert alpha == 1.

#     if N < 1:
#         return np.ones((0,2))

    ab = np.zeros((N+1, 2))

    ab[0, 1] = sp.gamma(1 + rho)
    ab[1:, 1] = np.arange(1, N+1) * (np.arange(1, N+1) + rho)
    ab[:, 1] = np.sqrt(ab[:, 1])

    ab[0, 0] = 0
    ab[1:, 0] = 2 * np.arange(N) + rho + 1

    return ab


def hfreud_idist_driver(x, n, alpha, rho, M=25):
    """
    Evaluates the integral, F = \\int_{0}^x p_n^2(x) \\dx{\\mu(x)} for x <= x0
    """
    if isinstance(x, float) or isinstance(x, int):
        x = np.asarray([x])
    else:
        x = np.asarray(x)

    F = np.zeros(x.size)

#     if alpha != 1:
#         HF = LaguerrePolynomials(alpha, rho)
#         x0 = HF.idist_medapprox(n)
#         rflags = x > x0
#         F[rflags] = 1 - hfreud_idistc_driver(x[rflags], n, alpha, rho, M)
#     else:
#         x0 = 50
#         rflags = x > x0
#         F[rflags] = 1 - hfreud_idistc_driver(x[rflags], n, alpha, rho, M)

    if alpha == 1:
        ab = laguerre_recurrence_values(n, alpha, rho)
        ab[0, 1] = 1
    if alpha == 2:
        ab = hermite_recurrence_values(n, rho/2)
        ab[0, 1] = 1

    if n > 0:
        xn, wn = gauss_quadrature_driver(ab, n)
        logfactor0 = -np.sum(np.log(ab[:, 1]**2))
    else:
        logfactor0 = 0
        xn = []

    ab_J = jacobi_recurrence_values(n+M+1, 0, rho)
    ab_J[0, 1] = 1

    for ind in range(x.size):
        if x[ind] == 0:
            F[ind] = 0
            continue

#         if rflags[ind]:
#             continue

        un = 2 * xn / x[ind] - 1
        ab = ab_J
        logfactor = logfactor0

        for j in range(n):
            ab = quadratic_modification(ab, un[j])
            logfactor = logfactor + np.log(ab[0, 1]**2)
            ab[0, 1] = 1.

        u, w = gauss_quadrature_driver(ab, M)

        Ival = np.sum(w * np.exp(- (x[ind]/2)**alpha * (u+1)**alpha))

        logfactor = logfactor + (2*n+rho+1) * np.log(x[ind]/2) + np.log(alpha) + \
                                (rho+1) * np.log(2) - sp.gammaln((rho+1)/alpha) - np.log(rho+1)

        F[ind] = np.exp(logfactor + np.log(Ival))

    return F


def hfreud_idistc_driver(x, n, alpha, rho, M=25):
    """
    Evaluates the integral, F = \\int_{0}^x p_n^2(x) \\dx{\\mu(x)} for x >= x0
    """
    if isinstance(x, float) or isinstance(x, int):
        x = np.asarray([x])
    else:
        x = np.asarray(x)

    F = np.zeros(x.size)

    if alpha == 1:
        ab = laguerre_recurrence_values(n, alpha, rho)
        ab[0, 1] = 1
    if alpha == 2:
        ab = hermite_recurrence_values(n, rho/2)
        ab[0, 1] = 1

    if n > 0:
        xn, wn = gauss_quadrature_driver(ab, n)
        logfactor0 = -np.sum(np.log(ab[:, 1]**2))
    else:
        logfactor0 = 0
        xn = []

    R = int(np.floor(np.abs(rho)))
    ab_H = laguerre_recurrence_values(n+M+R+1, alpha, 0)
    ab_H[0, 1] = 1

    for ind in range(x.size):
        if x[ind] == 0:
            F[ind] = 0
            continue

        un = xn - x[ind]

        ab = ab_H
        logfactor = logfactor0

        for j in range(n):
            ab = quadratic_modification(ab, un[j])
            logfactor = logfactor + np.log(ab[0, 1]**2)
            ab[0, 1] = 1

        root = -x[ind]
        for k in range(R):
            ab = linear_modification(ab, root)
            logfactor = logfactor + np.log(ab[0, 1]**2)
            ab[0, 1] = 1

        u, w = gauss_quadrature_driver(ab, M)
        Ival = np.sum(w * (u+x[ind])**(rho-R) * np.exp(u**alpha + x[ind]**alpha - (u+x[ind])**alpha))

        logfactor = logfactor + (-x[ind]**alpha) + sp.gammaln(1/alpha) - sp.gammaln((rho+1)/alpha)

        F[ind] = np.exp(logfactor + np.log(Ival))

    return F


def hfreud_idist_medapprox(n, alpha, rho):

    if n > 0:
        a = rho + 2*n + 2*np.sqrt(n**2 + n*rho)  # maxapprox
        a = a ** (1/alpha)
        a = a * np.exp(1/alpha * (np.log(np.sqrt(np.pi)) + sp.gammaln(alpha)
                                  - np.log(2) - sp.gammaln(alpha+1/2)))

        b = rho + 2*n - 2*np.sqrt(n**2 + n*rho)  # minapprox
        b = b ** (1/alpha)
        b = b * np.exp(1/alpha * (np.log(np.sqrt(np.pi)) + sp.gammaln(alpha)
                                  - np.log(2) - sp.gammaln(alpha+1/2)))
    else:
        a = gamma.ppf(1-1e-3, (rho+1)/alpha)
        a = a**(1/alpha)

        b = gamma.ppf(1e-3, (rho+1)/alpha)
        b = a**(1/alpha)

    return a, b


def hfreud_idist(x, n, alpha, rho):

    if isinstance(x, float) or isinstance(x, int):
        x = np.asarray([x])
    else:
        x = np.asarray(x)

    if alpha != 1:
        x0 = sum(hfreud_idist_medapprox(n, alpha, rho)) / 2
    else:
        x0 = 50

    F = np.zeros(x.size,)
    F[np.where(x <= x0)] = hfreud_idist_driver(x[np.where(x <= x0)], n, alpha, rho)
    F[np.where(x > x0)] = 1 - hfreud_idistc_driver(x[np.where(x > x0)], n, alpha, rho)

    return F


def hfreud_idistc(x, n, alpha, rho):

    return 1 - hfreud_idist(x, n, alpha, rho)

#     if isinstance(x, float) or isinstance(x, int):
#         x = np.asarray([x])
#     else:
#         x = np.asarray(x)
#
#     if alpha != 1:
#         x0 = sum( hfreud_idist_medapprox(n, alpha, rho) ) / 2
#     else:
#         x0 = 50
#
#     F = np.zeros(x.size,)
#     F[np.where(x<=x0)] = 1 - hfreud_idist_driver(x[np.where(x<=x0)], n, alpha, rho)
#     F[np.where(x>x0)] = hfreud_idistc_driver(x[np.where(x>x0)], n, alpha, rho)
#
#     return F


def hfreud_tolerance(n, alpha, rho, tol):
    assert tol > 0 and tol < 1
    x = hfreud_idist_medapprox(n, alpha, rho)[0]
    F = hfreud_idistc_driver(x, n, alpha, rho)

    while F > tol:
        x += 1
        F = hfreud_idistc_driver(x, n, alpha, rho)

    return x


def hfreud_idistinv(u, n, alpha, rho):

    eps = np.finfo(float).eps

    if isinstance(u, float) or isinstance(u, int):
        u = np.asarray([u])
    else:
        u = np.asarray(u)

    if isinstance(n, float) or isinstance(n, int):
        n = np.asarray([n])
    else:
        n = np.asarray(n)

    if n.size == 1:
        n = n[0]
        rhs = 1.2 * hfreud_idist_medapprox(n, alpha, rho)[0]

        U = max(u)
        if U == 1:
            rhs = hfreud_tolerance(n, alpha, rho, eps/10)
        else:
            if np.abs(U) < 100*eps:
                rhs = hfreud_tolerance(n, alpha, rho, 1 - 10*eps)
            else:
                rhs = hfreud_tolerance(n, alpha, rho, 1-U)

        supp = np.array([0, rhs])
        ab = laguerre_recurrence_values(2*n + max(100, n), alpha, rho)

        def primitive(xx):
            return hfreud_idist(xx, n, alpha, rho)

        x = idistinv_driver(u, n, primitive, ab, supp)
    else:
        nmax = np.amax(n)
        ind = np.digitize(n, np.arange(-0.5, 0.5+nmax+1e-8), right=False)
        ab = laguerre_recurrence_values(2*nmax + max(100, nmax), alpha, rho)

        assert n.size == u.size

        x = np.zeros(n.shape)
        for qq in range(nmax+1):
            rhs = 1.2 * hfreud_idist_medapprox(qq, alpha, rho)[0]

            U = max(u)
            if U == 1:
                rhs = hfreud_tolerance(qq, alpha, rho, eps/10)
            else:
                rhs = hfreud_tolerance(qq, alpha, rho, 1-U)

            supp = [0, rhs]

            flags = ind == qq+1

            def primitive(xx):
                return hfreud_idist(xx, qq, alpha, rho)

            x[flags] = idistinv_driver(u[flags], qq, primitive, ab, supp)

    return x


class LaguerrePolynomials(OrthogonalPolynomialBasis1D):
    def __init__(self, alpha=1., rho=0., **options):
        OrthogonalPolynomialBasis1D.__init__(self, **options)
        assert alpha > 0
        assert rho > -1
        self.alpha = alpha
        self.rho = rho

    def recurrence_driver(self, N):
        if self.alpha == 1.:
            ab = laguerre_recurrence_values(N, self.alpha, self.rho)
        else:
            raise ValueError('Only alpha=1 half-Freud recurrence coefficients have explicit formulas')

        if self.probability_measure and N > 0:
            ab[0, 1] = 1.

        return ab

    def idist_medapprox(self, n):

        alpha = self.alpha
        rho = self.rho

        return sum(hfreud_idist_medapprox(n, alpha, rho)) / 2

    def idist(self, x, n):

        alpha = self.alpha
        rho = self.rho

        return hfreud_idist(x, n, alpha, rho)

    def idistinv(self, u, n):

        alpha = self.alpha
        rho = self.rho

        return hfreud_idistinv(u, n, alpha, rho)


def discrete_chebyshev_recurrence_values(N, M):
    """
    Returns the first N+1 recurrence coefficients pairs for the Discrete
    Chebyshev measure, the N-point discrete uniform measure with equispaced
    support on [0,1].
    """

    assert M > 0, N < M

    if N < 1:
        ab = np.ones((1, 2))
        ab[0, 0] = 0.
        ab[0, 1] = 1.
        return ab

    ab = np.ones((N+1, 2))
    ab[0, 0] = 0
    ab[1:, 0] = 0.5

    n = np.arange(1, N+1, dtype=float)
    ab[1:, 1] = M/(2*(M-1)) * np.sqrt((1 - (n/M)**2) / (4 - (1/n**2)))

    return ab


def discrete_chebyshev_idistinv_helper(u, support, idist_evals):
    """
    Performs the binning and subscripting for the idistinv routine.
    """

    M = len(idist_evals)
    bin_edges = np.concatenate([np.array([0.]), idist_evals])
    bins = np.digitize(u, bin_edges, right=True)

    bins -= 1
    bins[bins < 0] = 0  # These points should correspond to u=0
    bins[bins >= M] = M-1  # These points should correspond to u=1

    return support[bins]


class DiscreteChebyshevPolynomials(OrthogonalPolynomialBasis1D):
    """
    Class for polynomials orthonormal on [0,1] with respect to an M-point
    discrete uniform measure with support equidistributed on the interval.
    """
    def __init__(self, M=2, domain=[0., 1.]):
        OrthogonalPolynomialBasis1D.__init__(self)
        assert M > 1
        self.M = M

        assert len(domain) == 2
        self.domain = np.array(domain).reshape([2, 1])
        self.standard_domain = np.array([0, 1]).reshape([2, 1])
        self.transform_to_standard = AffineTransform(domain=self.domain, image=self.standard_domain)
        self.standard_support = np.linspace(0, 1, M)
        self.support = self.transform_to_standard.mapinv(self.standard_support)
        self.standard_weights = 1/M*np.ones(M)

    def recurrence_driver(self, N):
        # Returns the first N+1 recurrence coefficient pairs for the
        # Discrete Chebyshev polynomial family.
        assert(N < self.M)
        ab = discrete_chebyshev_recurrence_values(N, self.M)
        if self.probability_measure and N > 0:
            ab[0, 1] = 1.

        return ab

    def eval(self, x, n, **options):

        return super().eval(self.transform_to_standard.map(x), n, **options)

    def idist(self, x, n, nugget=False):
        """
        Evalutes the order-n induced distribution at the locations x.

        Optionally, add a nugget to ensure correct computation on the
        support points.
        """

        assert n >= 0

        x_standard = self.transform_to_standard.map(to_numpy_array(x))
        if nugget:
            x_standard += 1e-3*(np.max(x) - np.min(x))

        bins = np.digitize(x_standard, self.standard_support, right=False)

        cumulative_weights = np.concatenate([np.array([0.]), np.cumsum(self.eval(self.support, n)**2)])/self.M

        return cumulative_weights[bins]

    def idistinv(self, u, n):
        """
        Computes the inverse order-n induced distribution at the locations
        u.
        """

        u = to_numpy_array(u)
        assert np.all(u >= 0), "Input u must contain numbers between 0 and 1"
        assert np.all(u <= 1), "Input u must contain numbers between 0 and 1"

        n = to_numpy_array(n)

        x = np.zeros(u.size)

        if n.size == 1:

            return discrete_chebyshev_idistinv_helper(u, self.support, self.idist(self.support, n[0], nugget=True))

        else:
            nmax = np.amax(n)
            ind = np.digitize(n, np.arange(-0.5, 0.5+nmax+1e-8), right=False)

            for i in range(nmax+1):
                flags = ind == i+1
                x[flags] = discrete_chebyshev_idistinv_helper(u[flags], self.support, self.idist(self.support, i, nugget=True))

            return x

    def fidistinv(self, u, n):
        """
        Fast routine for idistinv.
        (In this case, the "slow" routine is already very fast, so this
        is just an alias for idistinv.)
        """

        return self.idistinv(u, n)



from itertools import combinations

import numpy as np
from scipy import special as sp
from scipy.special import comb

from UncertainSCI.utils.prob import discrete_sampling, choice

def hyperbolic_cross_indices(d, k):
    """
    Returns indices associated with a d-dimensional (isotropic)
    hyperbolic cross index space up to degree k.
    """

    assert k >= 0
    assert d >= 1

    if d == 1:
        lambdas = range(k+1)
        return lambdas

    lambdas = np.zeros([1, d], dtype=int)

    # First add all indices with sparsity 1
    for q in range(d):
        temp = np.zeros([k-1, d], dtype=int)
        temp[:, q] = np.arange(1, k, dtype=int)
        lambdas = np.vstack([lambdas, temp])

    # Now determine the maximum 0-norm the entries can be. I.e., for
    # which values of p is 2^p <= k+1?
    pmax = int(np.floor(np.log(k+1)/np.log(2)))

    # For each sparsity p, populate with all possible indices of that
    # sparsity
    for p in range(2, pmax+1):
        # Determine all possible locations where nonzero entries can occur
        combs = combinations(range(d), p)
        combs = np.array([row for row in combs], dtype=int)

        # Now we have 2^p < k+1, i.e., an index with nonzero entries
        # np.ones([p 1]) is ok.
        # Keep incrementing these entries until product exceeds k+1
        possible_indices = np.ones([1, p])
        ind = 0

        while ind < possible_indices.shape[0]:
            # Add any possibilities that are children of
            # possible_indices[ind,:]

            lambd = possible_indices[ind, :]
            for q in range(p):
                temp = lambd.copy()
                temp[q] += 1
                if np.prod(temp+1) <= k+1:
                    possible_indices = np.vstack([possible_indices, temp])

            ind += 1

        possible_indices = np.vstack({tuple(row) for row in possible_indices})
        arow = lambdas.shape[0]
        lambdas = np.vstack([lambdas,
                             np.zeros([combs.shape[0]*possible_indices.shape[0],
                                       d],
                                      dtype=int)])

    # Now for each combination, we put in possible_indices
        for c in range(combs.shape[0]):
            i1 = arow
            i2 = arow + possible_indices.shape[0]

            lambdas[i1:i2, combs[c, :]] = possible_indices

            arow = i2

    return lambdas


def total_degree_indices(d, k):
    # Returns multi-indices associated with d-variate polynomials of
    # degree less than or equal to k. Each row is a multi-index, ordered
    # in total-degree-graded reverse lexicographic ordering.

    assert d > 0
    assert k >= 0

    if d == 1:
        return np.arange(k+1, dtype=int).reshape([k+1, 1])

    # total degree indices up to degree k in d-1 dimensions:
    lambdasd1 = total_degree_indices(d-1, k)
    # lambdasd1 should already be sorted by total degree, which is
    # assumed below

    lambdas = np.zeros([np.round(int(comb(d+k, d))), d], dtype=int)

    i0 = 0
    for qk in range(0, k+1):

        n = int(np.round(comb(d-1+(k-qk), d-1)))
        i1 = i0 + n

        lambdas[i0:i1, 0] = qk
        lambdas[i0:i1, 1:] = lambdasd1[:n, :]
        i0 = i1

    # My version of numpy < 1.12, so I don't have np.flip :(
    # degrees = np.cumsum(np.flip(lambdas,axis=1), axis=1)
    degrees = np.cumsum(np.fliplr(lambdas), axis=1)

    ind = np.lexsort(degrees.transpose())
    lambdas = lambdas[ind, :]

    return lambdas


def degree_encompassing_N(d, N):
    # Returns the smallest degree k such that nchoosek(d+k,d) >= N

    k = 0
    while np.round(comb(d+k, d)) < N:
        k += 1

    return k


def total_degree_indices_N(d, N):
    # Returns the first N ( > 0) d-dimensional multi-indices when ordered by
    # total degree graded reverse lexicographic ordering.

    assert N > 0

    return total_degree_indices(d, degree_encompassing_N(d, N))[:N, :]


def tensor_product(d, k):
    # Returns multi-indices associated with d-variate polynomials of
    # degree less than or equal to k. Each row is a multi-index, ordered
    # in p degree, where p = Inf

    from itertools import product

    Ival = np.empty(shape=[0, d], dtype=int)

    for t in (_ for _ in product(range(k+1), repeat=d)):
        Ival = np.vstack((Ival, np.asarray(t)))

    return Ival


def multi_indices_degree(d, k, p):
    # Returns multi-indices associated with d-variate polynomials of
    # degree less than or equal to k. Each row is a multi-index, ordered
    # in p degree, p could be any positive number including numpy.inf

    if p < 1:
        lambdas = total_degree_indices(d, k)
        norm = (np.sum(lambdas**p, axis=1))**(1/p)
        norm = np.round(norm, decimals=8)
        flags = (norm <= k)
        lambdas = lambdas[flags]

    elif p == np.inf:
        lambdas = tensor_product(d, k)

    elif p == 1:
        lambdas = total_degree_indices(d, k)

    else:
        lambdas = tensor_product(d, k)
        norm = (np.sum(lambdas**p, axis=1))**(1/p)
        norm = np.round(norm, decimals=8)
        flags = (norm <= k)
        lambdas = lambdas[flags]

    return lambdas


def pdjk(d, k):
    j = np.arange(k+1)
    p = np.exp(np.log(d) + sp.gammaln(k+1) - sp.gammaln(j+1) +
               sp.gammaln(j+d) - sp.gammaln(k+d+1))
    assert np.abs(sum(p)-1) < 1e-8
    return p


def sampling_total_degree_indices(N, d, k):

    """
    Chooses N random multi-indices (with the uniform probability law) from the
    set of d-variate multi-indices whose total degree is k and less

    Parameters
    ------
    param1: N
    Numebr of chosen random multi-indices
    param2: d
    dimension of variables
    param3L k
    total degree of variables

    Returns
    ------
    The output lambdas is an N x d matrix, with each row containing one of
    these multi-indices
    """
    lambdas = np.zeros((N, d))

    degrees = discrete_sampling(N, pdjk(d, k), np.arange(k+1)).T

    for i in range(1, d):
        for n in range(1, N+1):
            lambdas[n-1, i-1] = \
                discrete_sampling(1, pdjk(d-i, degrees[n-1]),
                                  np.arange(degrees[n-1], 0-1e-8, -1))

        degrees = degrees - lambdas[:, i-1]

    lambdas[:, d-1] = degrees

    return lambdas


class MultiIndexSet():
    def __init__(self, dim=None):
        self.dim = dim
        self.indices = np.zeros([0, self.dim])
        self.adaptive = False

    def get_indices(self):
        return self.indices

    def choice(self, n=1):
        """
        Returns n randomly chosen multi-indices as a (n x self.dim)
        numpy.ndarray.
        """

        assert n > 0

        index_choices = choice(range(self.indices.shape[0]), size=n)
        return self.indices[index_choices,:]

    def size(self):
        return self.indices.shape[0]

    def max_univariate_degree(self, dim=None):
        """
        Returns the maximum index value along a given dimension. If no
        dimension is specified, returns the maximum degree across all
        dimensions.

        Args:
            dim: A dimension integer taking values between 0 and self.dim-1,
              optional.
        Returns:
            deg: Maximum index value (integer).
        """

        if dim is None:
            return np.max(self.get_indices())
        else:
            return np.max(self.get_indices()[:, dim])

    def zero_indices(self, dim_indices=None):
        """
        Identifies indices in the index set whose entries in the dimensions
        dim_indices are 0.

        Args:
            dim_list: list-like iterable containing dimension indices.
        Returns:
            flags: A boolean numpy vector indicating which indices in the index
              set satisfy the conditions.
        """

        assert all([0 <= dim <= self.dim-1 for dim in dim_indices])

        if dim_indices is None:
            return self.get_indices()

        indices = self.get_indices()

        return np.linalg.norm(indices[:, dim_indices], axis=1) == 0

    def isamember(self, trial_indices):
        """
        Determines if input indices are members of the current index set.

        Args:
            trial_indices: An :math:`K \\times d` numpy array, where each row
              corresponds to an index.

        Returns:
            member: A numpy boolean array of size :math:`K` indicating if the
              rows of trial_indices are part of the current index set.
        """

        K, d = trial_indices.shape
        assert self.dim == d, \
               "Input index array should have {0:d} columns".format(self.dim)

        M = self.indices.shape[0]
        member = np.zeros(K, dtype=bool)
        for m in range(K):

            index = trial_indices[m, :]

            matches = np.ones(M, dtype=bool)
            for q in range(d):
                matches[matches] = index[q] == self.indices[matches, q]

            if np.any(matches):
                member[m] = True

        return member

    def get_margin(self):
        """
        Computes the margin of the index set :math:`\\Lambda`. In :math:`d`
        dimensions, this is defined as the set of indices :math:`\\lambda \\in
        N_0^d \\backslash \\Lambda` such that

        .. math::

          \\lambda - e_j \\in \\Lambda

        for some :math:`j = 1, \\ldots, d`.

        Returns:
            margin: A numpy array of size :math:`M \\times d` where each row
              contains an index in the margin.
        """

        # Do this in a brute-force manner:
        # - search for leaves of the current index set as margin candidates
        # - weed out leaves that are not in the margin

        M, d = self.indices.shape
        margin = np.zeros([0, d], dtype=self.indices.dtype)

        for m in range(M):
            candidates = np.tile(self.indices[m, :], [d, 1]) +\
                         np.eye(d, dtype=self.indices.dtype)
            membership_flags = ~self.isamember(candidates)
            if candidates[membership_flags, :].size > 0:
                margin = np.unique(
                          np.append(margin, candidates[membership_flags, :], axis=0),
                          axis=0)

        return margin

    def get_reduced_margin(self):
        """
        Computes the reduced margin of the index set :math:`\\Lambda`. In
        :math:`d` dimensions, this is defined as the set of indices
        :math:`\\lambda \\in N_0^d \\backslash \\Lambda` such that

        .. math::

          \\lambda - e_j \\in \\Lambda

        for every :math:`j = 1, \\ldots, d` satisfying :math:`\\lambda_j \\neq
        0`.

        Returns:
            margin: A numpy array of size :math:`M \\times d` where each row
              contains an index in the margin.
        """

        # We'll sequentially test elements in the margin

        margin = self.get_margin()
        K, d = margin.shape
        reduced_margin_inds = []

        for k in range(K):
            candidates = np.tile(margin[k, :], [d, 1]) - np.eye(d)
            candidates = candidates[~np.any(candidates < 0, axis=1), :]

            if np.all(self.isamember(candidates)):
                reduced_margin_inds.append(k)

        return margin[reduced_margin_inds, :]

    def augment(self, indices):
        """
        Augments the index set with the given indices.
        """

        K, d = indices.shape
        assert d == self.dim, \
               "Input index array should have {0:d} columns".format(self.dim)

        membership_flags = self.isamember(indices)
        if np.any(~membership_flags):
            self.adaptive = True
            self.indices = np.append(self.indices,
                                     indices[~membership_flags, :], axis=0)


class LpSet(MultiIndexSet):
    def __init__(self, dim=1, order=0, p=1):
        assert dim > 0 and order >= 0 and p >= 0

        super().__init__(dim=dim)
        self.dim = dim
        self.order = order
        self.p = p
        self.indices = self.get_indices()

    def get_indices(self):
        if self.p < 1:
            lambdas = total_degree_indices(self.dim, self.order)
            norm = (np.sum(lambdas**self.p, axis=1))**(1/self.p)
            norm = np.round(norm, decimals=8)
            flags = (norm <= self.order)
            lambdas = lambdas[flags]

        elif self.p == np.inf:
            lambdas = tensor_product(self.dim, self.order)

        elif self.p == 1:
            lambdas = total_degree_indices(self.dim, self.order)

        else:
            lambdas = tensor_product(self.dim, self.order)
            norm = (np.sum(lambdas**self.p, axis=1))**(1/self.p)
            norm = np.round(norm, decimals=8)
            flags = (norm <= self.order)
            lambdas = lambdas[flags]

        return lambdas


class TotalDegreeSet(MultiIndexSet):
    def __init__(self, dim=1, order=0):
        assert dim > 0 and order >= 0

        super().__init__(dim=dim)

        self.dim, self.order = dim, order
        self.indices = self.get_indices()

    def get_indices(self):
        if self.adaptive:
            return super().get_indices()
        else:
            return total_degree_indices(self.dim, self.order)


class HyperbolicCrossSet(MultiIndexSet):
    def __init__(self, dim=1, order=0):
        assert dim > 0 and order >= 0

        super().__init__(dim=dim)

        self.dim, self.order = dim, order
        self.indices = self.get_indices()

    def get_indices(self):
        if self.adaptive:
            return super().get_indices()
        else:
            return hyperbolic_cross_indices(self.dim, self.order)


if __name__ == "__main__":

    pass



import numpy as np
from scipy import sparse as sprs


class AffineTransform:
    def __init__(self, domain=None, image=None, A=None, b=None):
        """
        Initializes a general affine map of the form

          x ---> x*A.T + b,

        where x is a (row) vector, A is an invertible matrix, and b is a
        vector.

        If domain and image are specified, they are each 2 x d matrices
        specifying box constraints on a d-dimensional hypercube.
        """

        if domain is not None:
            if image is None:
                raise ValueError('If domain is specified, '
                                 'image must also be specified.')
            else:
                assert domain.shape == image.shape, \
                       "Domain, image matrices must be of same shape"

            self.diagonal = True
            d = domain.shape[1]
            self.A = sprs.eye(d)

            a = np.zeros(d)
            b = np.zeros(d)
            for q in range(d):
                a[q] = (image[1, q] - image[0, q]) / \
                       (domain[1, q] - domain[0, q])
                b[q] = image[0, q] - domain[0, q]*a[q]

            self.A = sprs.diags(a, 0)
            self.b = b

            self.Ainv = sprs.diags(1/a, 0)
            self.binv = self.Ainv.dot(-self.b)

        elif (A is not None) and (b is not None):
            # Assume A is a numpy array
            assert (A.ndim == 1) or (A.shape[0] == A.shape[1]), \
                   ValueError("Input matrix A must be square")
            assert b.ndim == 1, ValueError("Input b must be a vector")
            assert b.size == A.shape[0], \
                   ValueError("Input vector b must have same dimension as A")

            if A.ndim == 1:
                A = np.reshape(A, [1,1])

            self.A, self.b = A, b
            self.Ainv = np.linalg.inv(A)
            self.binv = self.Ainv.dot(-self.b)
        else:
            raise ValueError('Domain/image or A/b must be specified')

    def map(self, x):
        if len(x.shape) < 2:
            # either len(x) == dim or dim == 1
            if len(x) == self.b.size:
                return self.A.dot(x) + self.b
            elif self.b.size == 1:
                if isinstance(self.A, sprs.spmatrix):
                    return self.A.todense()[0, 0]*x + self.b
                else:
                    return self.A[0, 0]*x + self.b

        else:
            return self.A.dot(x.T).T + self.b

    def mapinv(self, x):
        if len(x.shape) < 2:
            # either len(x) == dim or dim == 1
            if len(x) == self.binv.size:
                return self.Ainv.dot(x) + self.binv
            elif self.b.size == 1:
                if isinstance(self.A, sprs.spmatrix):
                    return self.Ainv.todense()[0, 0]*x + self.binv
                else:
                    return self.Ainv[0, 0]*x + self.binv

        else:
            return self.Ainv.dot(x.T).T + self.binv

    def jacobian_determinant(self):
        """ Returns the Jacobian determinant of the map.

        Args:
            None
        Returns:
            jacdet: positive float, the absolute value of the map determinant
        """

        if isinstance(self.A, sprs.spmatrix):
            return np.abs(np.linalg.det(self.A.todense()))
        else:
            return np.abs(np.linalg.det(self.A))

    def compose(self, other):
        """ Returns composition of two AffineTransform's

        Args:
            other: An AffineTransform instance where the domain and range are
              the same dimension as for self.
       Returns:
            composition: AffineTransform instance, corresponding to the map
              self :math:`\\circ` other
        """

        A1 = np.asarray(self.A.todense())
        A2 = np.asarray(other.A.todense())

        assert A1.shape[0] == A2.shape[0]

        return AffineTransform(A=(A1 @ A2), b=(A1 @ other.b + self.b))



"""
Contains classes/methods for general univariate orthogonal polynomial families.
- evaluation
- gauss quadrature
- ratio evaluations
- linear/quadratic measure modifications
"""

import numpy as np
from scipy import special as sp
from numpy.linalg import eigh
from scipy import optimize
from scipy.special import gammaln


def eval_driver(x, n, d, ab):
    # Evaluates univariate orthonormal polynomials given their
    # three-term recurrence coefficients ab (a, b).
    #
    # Evaluates the d'th derivative. (Default = 0)
    #
    # Returns a numel(x) x numel(n) x numel(d) array.

    nmax = np.max(n)

    p = np.zeros(x.shape + (nmax+1,))
    xf = x.flatten()

    p[:, 0] = 1/ab[0, 1]

    if nmax > 0:
        p[:, 1] = 1/ab[1, 1] * ((xf - ab[1, 0])*p[:, 0])

    for j in range(2, nmax+1):
        p[:, j] = 1/ab[j, 1] * ((xf - ab[j, 0])*p[:, j-1] - ab[j-1, 1]*p[:, j-2])

    if type(d) == int:
        d = [d]
#        if d == 0:
#            return p[:, n.flatten()]
#        else:
#            d = [d]

    preturn = np.zeros([p.shape[0], n.size, len(d)])

    def assign_p_d(dval, parray):
        """
        Assigns dimension 2 of the nonlocal array preturn according to values
        in the derivative list d.
        """
        nonlocal preturn

        indlocations = [i for i, val in enumerate(d) if val == dval]
        for i in indlocations:
            preturn[:, :, i] = parray[:, n.flatten()]

    assign_p_d(0, p)

    for qd in range(1, max(d)+1):

        pd = np.zeros(p.shape)

        for qn in range(qd, nmax+1):
            if qn == qd:
                # The following is an over/underflow-resistant way to
                # compute ( qd! * kappa_{qd} ), where qd is the
                # derivative order and kappa_{qd} is the leading-order
                # coefficient of the degree-qd orthogonal polynomial.
                # The explicit formula for the lading coefficient of the
                # degree-qd orthonormal polynomial is prod(1/b[j]) for
                # j=0...qd.
                pd[:, qn] = np.exp(sp.gammaln(qd+1) - np.sum(np.log(ab[:(qd+1), 1])))
            else:
                pd[:, qn] = 1/ab[qn, 1] * ((xf - ab[qn, 0]) * pd[:, qn-1] - ab[qn-1, 1] * pd[:, qn-2] + qd*p[:, qn-1])

        assign_p_d(qd, pd)

        p = pd

    if len(d) == 1:
        return preturn.squeeze(axis=2)
    else:
        return preturn

def leading_coefficient_driver(N, ab):
    """
    Returns the leading coefficients for the first N polynomial basis elements.
    """
    assert N > 0
    return np.cumprod(1 / ab[:N, 1])

def ratio_driver(x, n, d, ab):
    """
    Evalutes ratios of orthonormal polynomials. These are given by

      r_n(x) = p_n(x) / p_{n-1}(x),  n >= 1

    The output is a x.size x n.size array.
    """
    nmax = np.max(n)

    r = np.zeros([x.size, nmax+1])
    xf = x.flatten()

    r[:, 0] = 1/ab[0, 1]
    if nmax > 0:
        r[:, 1] = 1/ab[1, 1] * (x - ab[1, 0])

    for j in range(2, nmax+1):
        r[:, j] = 1/ab[j, 1] * ((xf - ab[j, 0]) - ab[j-1, 1]/r[:, j-1])

    r = r[:, n.flatten()]

    if type(d) == int:
        if d == 0:
            return r
        else:
            raise NotImplementedError()
    else:
        raise NotImplementedError()


def s_driver(x, n, ab):
    """
    The output is a x.size x (n+1) array.

    s_n(x) = p_n(x) / sqrt(sum_{j=1}^{n-1} p_j^2(x)), n >= 0

    s_0(x) = p_0(x)

    s_1(x) = 1 / b_1 * (x - a_1)

    s_2(x) = 1 / (b_2 * sqrt(1+s_1^2)) * ((x - a_2)*s_1 - b_1)

    Need {a_k, b_k} k up to n
    """

    xf = x.flatten()
    nmax = np.max(n)

    s = np.zeros((xf.size, nmax+1))

    s[:, 0] = 1 / ab[0, 1]

    if nmax > 0:
        s[:, 1] = 1 / ab[1, 1] * (x - ab[1, 0])

    if nmax > 1:
        s[:, 2] = 1 / np.sqrt(1 + s[:, 1]**2) * ((x - ab[2, 0]) * s[:, 1] - ab[1, 1])
        s[:, 2] = s[:, 2] / ab[2, 1]

    for j in range(3, nmax+1):
        s[:, j] = 1 / np.sqrt(1 + s[:, j-1]**2) * \
                  ((x - ab[j, 0]) * s[:, j-1] - ab[j-1, 1] * s[:, j-2] / np.sqrt(1 + s[:, j-2]**2))
        s[:, j] = s[:, j] / ab[j, 1]

    return s[:, n.flatten()]


def jacobi_matrix_driver(ab, N):
    """
    Returns the N x N jacobi matrix associated to the input recurrence
    coefficients ab. (Requires ab.shape[0] >= N+1.)
    """

    return np.diag(ab[1:N, 1], k=1) + np.diag(ab[1:(N+1), 0], k=0) + np.diag(ab[1:N, 1], k=-1)


def gauss_quadrature_driver(ab, N):
    """
    Computes the N-point Gauss quadrature rule associated to the
    recurrence coefficients ab. (Requires ab.shape[0] >= N+1.)
    """

    from numpy.linalg import eigh

    if N > 0:
        lamb, v = eigh(jacobi_matrix_driver(ab, N))
        return lamb, ab[0, 1]**2 * v[0, :]**2
    else:
        return np.zeros(0), np.zeros(0)


def quadratic_modification(alphbet, z0):
    """
    The input is a single (N+1) x 2 array

    The output is a single (N) x 2 array
    """

    ab = np.zeros([alphbet.shape[0] - 1, 2])
    C = s_driver(z0, np.arange(alphbet.shape[0], dtype=int), alphbet)[0, :]

    temp = alphbet[1:, 1] * C[1:] * C[0:-1] / np.sqrt(1 + C[0:-1]**2)
    temp[0] = alphbet[1, 1] * C[1]

    acorrect = np.diff(temp)
    ab[1:, 0] = alphbet[2:, 0] + acorrect

    temp = 1 + C[:]**2
    bcorrect = temp[1:] / temp[0:-1]
    bcorrect[0] = (1 + C[1]**2) / C[0]**2
    ab[:, 1] = alphbet[1:, 1] * np.sqrt(bcorrect)

    return ab


def markov_stiltjies(u, n, ab, supp):

    """ Uses the Markov-Stiltjies inequalities to provide a bounding interval for x,
    the solution to F_n(x) = u

    Parameters
    ------
    param1: u
    given, u in [0,1]

    param2: n
    the order-n induced distribution function associated to the measure with
    three-term recurrrence coefficients a, b, having support on the real-line
    interval defined by the length-2 vector supp

    param3,4: a, b
    three-term recurrrence coefficients a, b

    param5: supp
    support on the real-line interval defined by the length-2 vector supp


    Returns
    ------
    intervals: an (M x 2) matrix if u is a length-M vector

    Requires
    ------
    a.size >> n

    """

    assert type(n) is int

    x, v = gauss_quadrature_driver(ab, n)

    ab[0, 1] = 1

    for j in range(n):
        ab = quadratic_modification(ab, x[j])
        ab[0, 1] = 1

    N = ab.shape[0] - 1

    y, w = gauss_quadrature_driver(ab, N)

    if supp[1] > y[-1]:
        X = np.insert(y, [0, y.size], [supp[0], supp[1]])
        W = np.insert(np.cumsum(w), 0, 0)

    else:
        X = np.insert(y, [0, y.size], [supp[0], y[-1]])
        W = np.insert(np.cumsum(w), 0, 0)

    W = W / W[-1]

    W[np.where(W > 1)] = 1  # Just in case for machine eps issues
    W[-1] = 1

    if isinstance(u, float) or isinstance(u, int):
        u = np.asarray([u])
    else:
        u = np.asarray(u)

    j = np.digitize(u, W, right=False)  # bins[i-1] <= x < bins[i], left bin end is open
    jleft = j - 1
    jright = j + 1

    flags = j == N + 1
    jleft[flags] = N + 1
    jright[flags] = N + 1

    intervals = np.array([X[jleft], X[jright]])

    return intervals.T


def idistinv_driver(u, n, primitive, ab, supp):

    """
    Uses bisection to compute the (approximate) inverse of the order-n induced
    primitive function F_n

    Parameters
    ------
    param3: primitive
    The input function primitive should be a function handle accepting a single input
    and outputs the primitive F_n evaluated at the input

    Returns
    ------
    The ouptut x = F_n^{-1}(u)

    """

    if isinstance(u, float) or isinstance(u, int):
        u = np.asarray([u])
    else:
        u = np.asarray(u)

    if isinstance(n, np.int64):
        intervals = markov_stiltjies(u, int(n), ab, supp)
    elif isinstance(n, int):
        intervals = markov_stiltjies(u, n, ab, supp)
    else:
        intervals = np.zeros((n.size, 2))
        nmax = max(n)
        ind = np.digitize(n, np.arange(-0.5, 0.5+nmax+1e-8), right=False)
        for i in range(nmax+1):
            flags = ind == i+1
            intervals[flags, :] = markov_stiltjies(u[flags], i, ab, supp)

    x = np.zeros(u.size,)
    for j in range(u.size):
        x[j] = optimize.bisect(lambda xx: primitive(xx) - u[j], intervals[j, 0], intervals[j, 1])

    return x


def linear_modification(alphbet, x0):
    """
    The input is a single (N+1) x 2 array

    The output is a single N x 2 array

    The appropriate sign of the modification (+/- (x-x0)) is inferred from the
    sign of (alph(1) - x0). Since alph(1) is the zero of p_1, then it is in
    \\supp \\mu
    """
    sgn = np.sign(alphbet[1, 0] - x0)

    ns = np.arange(alphbet.shape[0], dtype=int)
    r = np.abs(ratio_driver(x0, ns, 0, alphbet)[0, 1:])
    assert r.size == alphbet.shape[0] - 1

    ab = np.zeros([alphbet.shape[0]-1, 2])

    acorrect = alphbet[1:-1, 1] / r[:-1]
    acorrect[1:] = np.diff(acorrect)
    ab[1:, 0] = alphbet[1:-1, 0] + sgn * acorrect

    bcorrect = alphbet[1:, 1] * r
    bcorrect[1:] = bcorrect[1:] / bcorrect[:-1]
    ab[:, 1] = alphbet[:-1, 1] * np.sqrt(bcorrect)
    
    return ab


def derivative_expansion_driver(ab, s, N, K):
    """
    Computes the coefficients 

    .. math::

      \\sigma^{(s)}_{n,k} = \\left\\langle p_n^{(s)}, p_k \\right\\rangle,

    where :math:`p_n^{(s)}` is the :math:`s`th derivative of the 
    degree-:math:`k` orthonormal polynomial :math:`p_k`. These are,
    equivalently, expansion coefficients of :math:`p_n^{(s)}` in the basis
    :math:`\\{p_k\\}_k`.

    Args:
        ab: The recurrence coefficients for the family. An Mx2 numpy array,
          where M must be at least max(N+1, K+1).
        s: The integer order of the derivative. Must be non-negative.
        N: Computes coefficients for :math:`n \\leq N`
        K: Computes coefficients for :math:`k \\leq K`.
    Returns:
        C: (N+1) x (K+1) numpy array containing coefficients.
    """

    assert ab.shape[1] == 2

    M = ab.shape[0]
    NK = max(N, K)

    assert M >= NK+1

    if s==0:
        return np.eye(NK+1)[:(N+1), :(K+1)]
    if N < s:
        return np.zeros([N+1, K+1])

    # s=0 coefficients
    Cp = np.eye(NK+1)
    C = np.zeros([NK+1, NK+1])

    # Iterate over values of s
    for q in range(1, s+1):

        # Explicit starting value
        C[q,0] = np.exp(gammaln(q+1) - np.sum(np.log(ab[1:(q+1), 1])))

        for n in range(q+1, N+1):
            # k=0 is special
            C[n, 0] = q*Cp[n-1, 0] + (ab[1, 0] - ab[n, 0])*C[n-1, 0] - \
                     ab[n-1, 1]*C[n-2, 0] + ab[1, 1]*C[n-1, 1]
            C[n, 0] /= ab[n, 1]

            for k in range(1, n-q+1):
                C[n, k] = q*Cp[n-1, k] + (ab[k+1, 0] - ab[n, 0])*C[n-1, k] - \
                          ab[n-1, 1]*C[n-2, k] + ab[k+1, 1]*C[n-1, k+1] + \
                          ab[k, 1]*C[n-1, k-1]
                C[n, k] /= ab[n, 1]

        Cp = C.copy()
        C = np.zeros([NK+1, NK+1])

    return Cp[:(N+1),:(K+1)]

class OrthogonalPolynomialBasis1D:
    def __init__(self, recurrence=[], probability_measure=True):
        self.probability_measure = probability_measure
        self.ab = np.zeros([0, 2])
        pass

    def recurrence(self, N):
        """
        Returns the first N+1 orthogonal polynomial recurrence pairs.
        The orthonormal polynomial family satisfies the recurrence

          p_{-1}(x) = 0
          p_0(x) = 1/ab[0,1]

          x p_n(x) = ab[n+1,1] p_{n+1}(x) + ab[n+1,0] p_n(x) + ab[n,1] p_{n-1}(x)
           (n >= 0)

        The value ab[0,0] is ignored and never used.

        Recurrence coefficients ab, once computed, are stored as an
        instance variable. On subsequent calls to this function, the
        stored values are returned if the instance variable already
        contains the desired coefficients. If the instance variable does
        not contain enough coefficients, then a call to
        recurrence_driver is performed to compute the desired
        coefficients, and the output is stored in the instance variable.

        Parameters
        ----------
        N: positive integer
            Maximum polynomial degree for desired recurrence coefficients

        Returns
        -------
        ab: ndarray
            (N+1) x 2 array of recurrence coefficients.
        """

        if N+1 > self.ab.shape[0]:
            self.ab = self.recurrence_driver(N)
            return self.ab
        else:
            return self.ab[:(N+1), :]

    # Recurrence coefficient functions should be defined as follows:
    # The returned array has size (N+1) x 2. The [0,0] entry is not used
    # and set to 0. If the array is ab, then the orthonormal version of
    # the three-term recurrence with the degree-n polynomial p_n at x is:
    #
    #   ab[n+1, 1] * p_{n+1} = (x - ab[n+1,0]) * p_n - ab[n, 1] p_{n-1}
    def recurrence_driver(self, N):
        raise ValueError('Define this')
        return

    def eval(self, x, n, d=0):
        # Evaluates univariate orthonormal polynomials given their
        # three-term recurrence coefficients ab.
        #
        # Evaluates the d'th derivative. (Default = 0)
        #
        # Returns a numel(x) x numel(n) x numel(d) array.

        n = np.asarray(n)
        if isinstance(x, int) or isinstance(x, float):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        if n.size < 1 or x.size < 1:
            return np.zeros(0)

        nmax = np.max(n)
        ab = self.recurrence(nmax)

        assert nmax < ab.shape[0]
        assert np.min(n) > -1
        assert np.all(d >= 0)

        return eval_driver(x, n, d, ab)

    def jacobi_matrix_driver(ab, N):
        """
        Returns the N x N jacobi matrix associated to the input recurrence
        coefficients ab. (Requires ab.shape[0] >= N+1.)
        """

        return np.diag(ab[1:N, 1], k=1) + np.diag(ab[1:(N+1), 0], k=0) + np.diag(ab[1:N, 1], k=-1)

    def jacobi_matrix(self, N):
        """
        Returns the N x N jacobi matrix associated to the polynomial family.
        """
        return jacobi_matrix_driver(self.recurrence(N+1), N)
        # J = np.diag(ab[1:N,1], k=1) + np.diag(ab[1:(N+1),0],k=0) + np.diag(ab[1:N,1], k=-1)
        # return J

    def apply_jacobi_matrix(self, v):
        """
        Premultiplies the input array by the Jacobi matrix of the
        appropriate size for the polynomial family. Applies the Jacobi
        matrix across the first dimension of v.

        Parameters
        ----------
        v: ndarray
            Input vector or array

        Returns
        -------
        Jv: ndarray
            J*v, where J is the Jacobi matrix of size v.shape[0].
        """

        N = v.shape[0]
        ab = self.recurrence(N+1)

        # Rebroadcast v so we can take advantage of numpy's
        # multiplication
        v = np.moveaxis(v, 0, -1)

        # J = np.diag(ab[1:N,1], k=1) + np.diag(ab[1:(N+1),0],k=0) + np.diag(ab[1:N,1], k=-1)
        Jv = v*ab[1:(N+1), 0]
        Jv[..., :-1] += v[..., 1:]*ab[1:N, 1]
        Jv[..., 1:] += v[..., :-1]*ab[1:N, 1]
        Jv = np.moveaxis(Jv, -1, 0)

        return Jv

    def gauss_quadrature(self, N):
        """
        Computes the N-point Gauss quadrature rule associated to the
        recurrence coefficients ab.
        """

        return gauss_quadrature_driver(self.recurrence(N+1), N)

    def gauss_radau_quadrature(self, N, anchor=0.):
        """
        Computes the N-point Gauss quadrature rule associated to the
        polynomial family, with a node at the specified anchor.
        """

        # ## Note: the quadrature weight underflows for anchor far
        # ## outside the support interval. This causes imprecise quadrature
        # ## results for large-degree polynomials evaluated far outside
        # ## the support interval.

        ab = self.recurrence(N+1)
        c = self.r_eval(anchor, N)

        cd = ab.copy()
        cd[N, 0] += c*cd[N, 1]

        J = np.diag(cd[1:N, 1], k=1) + np.diag(cd[1:(N+1), 0], k=0) + np.diag(cd[1:N, 1], k=-1)
        lamb, v = eigh(J)

        return lamb, v[0, :]**2

    def leading_coefficient(self, N):
        """
        Returns the leading coefficients for the first N polynomial basis elements.
        """
        assert N > 0
        return np.cumprod(1 / self.recurrence(N)[:, 1])

    def canonical_connection(self, N):
        """
        Returns the N x N matrix C, where row n of C contains expansion
        coefficients for p_n in the monomial basis. I.e.,

         p_n(x) = sum_{j=0}^{n} C[n,j] x**j,

        for n = 0, ..., N-1.
        """

        ab = self.recurrence(N)
        C = np.zeros([N, N])

        if N < 1:
            return C

        C[0, 0] = 1/ab[0, 1]
        if N == 1:
            return C

        C[1, 1] = C[0, 0]/ab[1, 1]
        C[1, 0] = -ab[1, 0]*C[0, 0]/ab[1, 1]

        for n in range(1, N-1):
            C[n+1, 0] = -ab[n+1, 0]*C[n, 0] - ab[n, 1]*C[n-1, 0]
            C[n+1, n] = C[n, n-1] - ab[n+1, 0]*C[n, n]
            C[n+1, n+1] = C[n, n]

            js = np.arange(1, n)
            C[n+1, js] = C[n, js-1] - ab[n+1, 0]*C[n, js] - ab[n, 1]*C[n-1, js]

            C[n+1, :] /= ab[n+1, 1]

        return C

    def canonical_connection_inverse(self, N):
        """
        Returns the N x N matrix C, where row n of C contains expansion
        coefficients for x^n in the orthonormal basis . I.e.,

         x^n = sum_{j=0}^{n} C[n,j] p_j(x)

        for n = 0, ..., N-1.
        """

        ab = self.recurrence(N)
        C = np.zeros([N, N])

        if N < 1:
            return C

        C[0, 0] = ab[0, 1]
        if N == 1:
            return C

        C[1, 1] = ab[0, 1]*ab[1, 1]
        C[1, 0] = ab[1, 0]*ab[0, 1]

        for n in range(1, N-1):
            C[n+1, :] = self.apply_jacobi_matrix(C[n, :])

        return C

    def tuple_product_generator(self, IC, ab=None):
        """
        Helper function that increments indices for a polynomial product expansion.

        IC is a vector with entries

          IC[j] = < p_j, p_alpha >, j \\in range(N),

        where N = IC.size. The notation < ., .> is the inner product
        under which the polynomial family is orthonormal. alpha is a
        multi-index of arbitrary shape with a polynomial defined by

          p_alpha = \\prod_{j \\in range(alpha.size)} p_{alpha[j]}(x).

        The value of alpha is not needed by this function.

        This function returns an N x N matrix C with entries

            C[n,j] = < p_n p_j, p_alpha >, j, n \\in range(N)

        Parameters
        ----------
        IC: vector (1d array)
            Values of input inner products

        ab: ndarray, optional
            Recurrence coefficients

        Returns
        -------
        C: ndarray
            Output coefficient expansion vector for beta
        """

        N = IC.size
        ab = self.recurrence(N+1)
        C = np.zeros((N, N))
        C[0, :] = IC

        for n in range(N-1):
            C[n+1, :] = self.apply_jacobi_matrix(C[n, :]) - ab[n+1, 0]*C[n, :]
            if n > 0:
                C[n+1, :] -= ab[n, 1]*C[n-1, :]
            C[n+1, :] /= ab[n+1, 1]

        return C

    def tuple_product(self, N, alpha):
        """
        Computes integrals of polynomial products. Returns an N x N matrix C with entries

            C[n,m] = < p_n p_m, p_alpha >,

        where alpha is a vector of integers and p_alpha is defined

            p_alpha = prod_{j=1}^{alpha.size} p_[alpha[j]](x),

        The notation <., .> denotes the inner product under which the
        polynomial family is orthogonal.

        Parameters
        ----------
        N: integer
            Size of matrix to return

        alpha: ndarray (1d)
            Multi-index defining a polynomial product

        Returns
        -------
        C: ndarray
            Output N x N matrix containing integral values
        """

        M = alpha.size
        if M == 0:
            return np.eye(N)

        # Nmax = max(N, np.max(alpha)+1)
        Nmax = N + np.sum(alpha) + 1
        C = np.zeros((Nmax, Nmax))

        ab = self.recurrence(Nmax+1)

        # C[j,k] = delta_{j,k}
        # Initial condition: IC[j] = < p_0 p_j, p_{alpha[0]} >
        #                          = C[alpha[0],:]
        C[alpha[0], alpha[0]] = 1.

        for j in range(M):
            IC = C[alpha[j], :]/ab[0, 1]
            C = self.tuple_product_generator(IC, ab=ab)

        return C[:N, :N]

    def derivative_expansion(self, N, d):
        """
        Computes an N x N matrix with expansion coefficients for
        derivatives of orthogonal polynomials. I.e., computes the numbers C[n,j] for

          p_n^{(d)}(x) = sum_{j=0}^n C[n,j] p_j(x),

        for j,n = 0, ..., N-1.

        Parameters
        ----------
        N: integer
            Size of matrix to return

        d: integer
            Derivative order

        Returns
        -------
        C: ndarray
            Output N x N matrix containing expansion coefficients
        """

        assert N >= 0
        assert d >= 0

        if N == 0:
            return np.zeros(0)

        if d == 0:
            return np.eye(N)

        ab = self.recurrence(N+1)
        C = np.eye(N+d+1)[:N, :]

        for dj in range(1, d+1):
            Cprev = C[:, :-1].copy()
            C = np.zeros([N, N+d+1-dj])
            C[dj, 0] = np.exp(gammaln(dj+1) - np.sum(np.log(ab[1:(dj+1), 1])))

            for n in range(dj, N-1):
                C[n+1, :] = 1/ab[n+1, 1] * (self.apply_jacobi_matrix(C[n, :])
                                            - ab[n+1, 0]*C[n, :] - ab[n, 1]*C[n-1, :] + dj*Cprev[n, :])

        return C[:, :-1]

    def r_eval(self, x, n, d=0):
        """
        Evalutes ratios of orthonormal polynomials. These are given by

          r_n(x) = p_n(x) / p_{n-1}(x),  n >= 1

        The output is a x.size x n.size array.
        """

        n = np.asarray(n)

        if isinstance(x, float) or isinstance(x, int):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        if n.size < 1 or x.size < 1:
            return np.zeros(0)

        nmax = np.max(n)
        ab = self.recurrence(nmax+1)

        assert nmax < ab.shape[0]
        assert np.min(n) > -1
        assert np.all(d >= 0) and np.all(d < 1)

        return ratio_driver(x, n, d, ab)

    def s_eval(self, x, n):
        """
        The output is a x.size x (n+1) array.

        s_n(x) = p_n(x) / sqrt(sum_{j=0}^{n-1} p_j^2(x)), n >= 0

        s_0(x) = p_0(x)

        s_1(x) = 1 / b_1 * (x - a_1)

        s_2(x) = 1 / (b_2 * sqrt(1+s_1^2)) * ((x - a_2)*s_1 - b_1)

        Need {a_k, b_k} k up to n
        """

        n = np.asarray(n)

        if isinstance(x, float) or isinstance(x, int):
            x = np.asarray([x])
        else:
            x = np.asarray(x)

        nmax = np.max(n)
        ab = self.recurrence(nmax+1)

        assert nmax < ab.shape[0]
        assert np.min(n) > -1

        return s_driver(x, n, ab)

    def qpoly1d_eval(self, x, n, d=0):
        """
        Evalutes Christoffel-function normalized polynomials. These are
        given by

          q_k(x) = p_k(x) / sqrt( sum_{j=0}^{n-1} p_j^2 ), k = 0, ..., n-1

        The output is a x.size x n array
        """

        assert n > 0

        ab = self.recurrence(n-1)

        q = np.zeros((x.size, n))
        q[:, 0] = 1.
        qt = np.zeros(x.size)

        if n > 1:
            qt = 1/ab[1, 1] * (x - ab[1, 0]) * q[:, 0]
            q[:, 1] = qt / np.sqrt(1 + qt**2)

        for j in range(1, n-1):
            qt = 1/ab[j+1, 1] * ((x - ab[j+1, 0])*q[:, j] - ab[j, 1] * q[:, j-1] / np.sqrt(1 + qt**2))
            q[:, j+1] = qt / np.sqrt(1 + qt**2)

        if type(d) == int:
            if d == 0:
                return q
            else:
                d = [d]
                raise NotImplementedError()

        raise NotImplementedError()

        qreturn = np.zeros([q.shape[0], q.shape[1], len(d)])
        for (qi, qval) in enumerate(d):
            if qval == 0:
                qreturn[:, :, qi] = q

        for qd in range(1, max(d)+1):
            assert False

        return qreturn

    def christoffel_function(self, x, k):
        """
        Computes the normalized (inverse) Christoffel function lambda, defined as

          lambda**2 = k / sum(p**2, axi=1),

        where p is a matrix containing evaluations of an orthonormal
        polynomial family up to degree k-1, defined by the recurrence
        coefficients ab.
        """

        assert k > 0

        p = self.eval(x, range(k))
        return np.sqrt(float(k) / np.sum(p**2, axis=1))

    def derivative_expansion(self, s, N, K=None):
        """
        Computes the coefficients 

        .. math::

          \\sigma^{(s)}_{n,k} = \\left\\langle p_n^{(s)}, p_k \\right\\rangle,

        where :math:`p_n^{(s)}` is the :math:`s`th derivative of the 
        degree-:math:`k` orthonormal polynomial :math:`p_k`. These are,
        equivalently, expansion coefficients of :math:`p_n^{(s)}` in the basis
        :math:`\\{p_k\\}_k`.

        Args:
            s: The integer order of the derivative. Must be non-negative.
            N: Computes coefficients for :math:`n \\leq N`
            K: Computes coefficients for :math:`k \\leq K` (optional). If not
              given, is set to N.
        Returns:
            C: (N+1) x (K+1) numpy array containing coefficients.
        """


        if K is None:
            K = N

        assert N > -1 and K > -1 

        ab = self.recurrence(max(N,K))
        return derivative_expansion_driver(ab, s, N, K=K)


if __name__ == "__main__":

    import pdb
    from families import JacobiPolynomials

    J = JacobiPolynomials()
    N = 10
    K = 10
    s = 3
    ab = J.recurrence(max(N,K))

    x, w = J.gauss_quadrature(2*N)
    Vd = J.eval(x, range(N+1), d=s)
    V = J.eval(x, range(K+1))

    C = J.derivative_expansion(s, N, K)

    C2 = Vd.T @ np.diag(w) @ V
    C2[np.abs(C2)<1e-8] = 0



import numpy as np


def array_unique(a, tol=1e-8):
    """
    Given an 1d numpy array, selects eps-close unique elements and delete, then
    sort.
    """
    a_sort = np.sort(a)
    seq_diff = a_sort[1:] - a_sort[:-1]
    ind = np.where(seq_diff < tol)
    return np.delete(a_sort, ind)


if __name__ == '__main__':
    x = np.array([1, 2, 1-1e-12, 2-1e-4])
    y = array_unique(x)
    print(x, y)



import numpy as np


def discrete_sampling(N, probs, states):
    """ samples iid from a discrete probability measure
    Parameters
    ------
    param1: N
    Numeber of iid samples generated from a random variable X

    param2: probs
    X's probability mass function, prob(X = states[j]) = prob[j]

    param3: states
    values of random variable X, X = statesj[]

    Returns
    ------
    N samples iid from a discrete probability measure
    """

    p = probs[:] / sum(probs[:])

    j = np.digitize(np.random.random(N), np.cumsum(p), right=False)

    assert probs.size == states.size

    x = states[j]

    return x

# TODO: use numpy.choice, scipy.stats.discrete_rv
def choice(a, size=1, p=None):
    if p is None:
        p = np.ones(len(a))/len(a)
    return discrete_sampling(size, p, a)

if __name__ == "__main__":

    from families import JacobiPolynomials

    alph = -0.8
    bet = np.sqrt(101)
    states = np.linspace(-1, 1, 10)
    n = 4
    J = JacobiPolynomials(alpha=alph, beta=bet)

    probs = J.idist(states, n, M=10)

    N = 5
    print(discrete_sampling(N, probs, states))



import numpy as np
from numpy import random

from UncertainSCI.indexing import MultiIndexSet

def mixture_tensor_discrete_sampling(M, grids, weights, iset):
    """
    Constructs M random samples from a mixture of tensorial discrete measures.
    The measure is a uniform mixture of measures, each supported on the tensor
    product of grids (a list of arrays).

    The input weights define a collection of tensorial measures on the discrete
    support.

    The MultiIndexSet iset defines which measures are used in the mixture.

    weights is a list of arrays, where weights[j][:,i] corresponds to the
    i'th distribution on dimension j.
    """

    assert len(grids) == len(weights) == iset.dim

    N = [grids[j].size for j in range(iset.dim)]

    x = np.zeros([M, iset.dim])
    indices = iset.choice(M)

    max_indices = iset.maximum_unvivariate_degree()

    # Loop over possible values for the indices for each dimension.
    # This should be faster than looping over samples.

    for d in range(iset.dim):
        for k in range(max_indices[d]):
            mask = indices[:,d] == k
            x[mask,d] = random.choice(grids[d], probabilities=weights[d][:,k])

    return x



import warnings
from itertools import chain, combinations
from math import floor

import numpy as np
from scipy.stats.mstats import mquantiles

from UncertainSCI.indexing import MultiIndexSet, TotalDegreeSet
from UncertainSCI.distributions import ProbabilityDistribution, TensorialDistribution
from UncertainSCI.utils.casting import to_numpy_array
from UncertainSCI.utils.version import version_lessthan
from UncertainSCI.utils.linalg import lstsq_loocv_error, weighted_lsq
from UncertainSCI.sampling import mixture_tensor_discrete_sampling

valid_training_types = ['christoffel_lsq', 'wlsq']
valid_sampling_types = ['induced', 'gq', 'gq-induced', 'greedy-induced', 'greedy-gq-induced']

sampling_options = {'induced': [],
                    'gq': ['M'],
                    'gq-induced': ['M'],
                    'greedy-induced': ['oversampling'],
                    'greedy-gq-induced': ['M']}

training_options = {'wlsq': []}

class PolynomialChaosExpansion():
    """Base polynomial chaos expansion class.

    Provides interface to construct and manipulate polynomial chaos expansions.

    Attributes:
    -----------
        coefficients: A numpy array of polynomial chaos expansion coefficients.
        indices: A MultiIndexSet instance specifying the polynomial
          approximation space.
        distribution: A ProbabilityDistribution instance indicating the
          distribution of the random variable.
        samples: The experimental or sample design in stochastic space.

    """
    def __init__(self, index_set=None, distribution=None, order=None,
                       plabels=None, sampling='greedy-induced',
                       training='wlsq', **kwargs):

        self.coefficients = None
        self.accuracy_metrics = {}
        self.samples = None
        self.model_output = None

        self.sampling = sampling.lower()
        self.training = training.lower()

        if distribution is None:
            raise ValueError('A distribution must be specified')
        elif issubclass(type(distribution), ProbabilityDistribution):
            self.distribution = distribution
        else:
            try:
                iter(distribution)
                self.distribution = TensorialDistribution(distribution)
            except:
                raise ValueError('Invalid distribution specification')

        if plabels is None:
            self.plabels = [str(val+1) for val in range(self.distribution.dim)]
        else:
            if len(plabels) != self.distribution.dim:
                raise ValueError('Parameter labels input "plabels" must have \
                                  same size as number of parameters in given \
                                  input "distribution"')
            self.plabels = plabels

        if index_set is None:
            if order is None:
                raise ValueError('Either "index_set" or "order" must be specified')
            else:
                self.index_set = TotalDegreeSet(dim=self.distribution.dim, order=order)
        else:
            if order is not None:
                warnings.warn("Both inputs 'order' and 'index_set' specified. Ignoring 'order'.")
            self.index_set = index_set

        self.sampling_options = {key: kwargs[key] \
                                 for key in sampling_options[self.sampling] \
                                 if key in kwargs.keys()}
        self.training_options = {key: kwargs[key] \
                                 for key in training_options[self.training] \
                                 if key in kwargs.keys()}

    def check_training(self):
        """Determines if a valid training type has been specified."""

        if self.training.lower() not in valid_training_types:
            raise ValueError('Invalid training type specified')

    def check_sampling(self):
        """Determines if a valid sampling type has been specified."""

        if self.sampling.lower() not in valid_sampling_types:
            raise ValueError('Invalid sampling type specified')

    def set_indices(self, index_set):
        """Sets multi-index set for polynomial approximation.

        Args:
            indices: A MultiIndexSet instance specifying the polynomial
              approximation space.
        Returns:
            None:
        """
        if isinstance(index_set, MultiIndexSet):
            self.index_set = index_set
        else:
            raise ValueError('Indices must be a MultiIndexSet object')

    def set_distribution(self, distribution):
        """Sets type of probability distribution of random variable.

        Args:
            distribution: A ProbabilityDistribution instance specifying the
              distribution of the random variable.
        Returns:
            None:
        """

        if isinstance(distribution, ProbabilityDistribution):
            self.distribution = distribution
        else:
            raise ValueError(('Distribution must be a ProbabilityDistribution'
                              'object'))

    def check_distribution(self):
        if self.distribution is None:
            raise ValueError('First set distribution with set_distribution')

    def check_indices(self):
        if self.index_set is None:
            raise ValueError('First set indices with set_indices')

    def set_samples(self, samples):
        if samples.shape[1] != self.index_set.get_indices().shape[1]:
                raise ValueError('Input parameter samples '
                                 'have wrong dimension')

        self.samples = samples

    def map_to_standard_space(self, q):
        """Maps parameter values from model space to standard space.

        Parameters:
            q (array-like): Samples in model space.

        Returns:
            p (numpy.ndarray): Samples in standard space
        """

        q = np.asarray(q)
        return self.distribution.transform_standard_dist_to_poly.map(
                self.distribution.transform_to_standard.map(q))


    def map_to_model_space(self, p):
        """Maps parameter values from standard space to model space.

        Parameters:
            p (array-like): Samples in standard space.

        Returns:
            q (numpy.ndarray): Samples in model space
        """

        p = np.asarray(p)
        return self.distribution.transform_to_standard.mapinv(
                self.distribution.transform_standard_dist_to_poly.mapinv(p))

    def generate_samples(self, **kwargs):
        """Generates sample/experimental design for use in PCE construction.

        Parameters:
            new_samples (array-like, optional): Specifies samples that must be
                part of the ensemble.
        """

        self.check_distribution()
        self.check_indices()
        self.check_sampling()

        if 'new_samples' in kwargs.keys():
            new_samples = kwargs['new_samples']
        else:
            new_samples = None

        for key in kwargs.keys():
            if key in sampling_options[self.sampling]:
                self.sampling_options[key] = kwargs[key]

        if self.sampling.lower() == 'greedy-induced':
            if new_samples is None:
                p_standard = self.distribution.polys.wafp_sampling(
                               self.index_set.get_indices(), **self.sampling_options)

                # Maps to domain
                self.samples = self.map_to_model_space(p_standard)

            else:  # Add new_samples random samples
                x = self.map_to_standard_space(self.samples)

                x = self.distribution.polys.wafp_sampling_restart(
                        self.index_set.get_indices(), x, new_samples,
                        **self.sampling_options)

                self.samples = self.map_to_model_space(x)

            self.weights = self.christoffel_weights()

        elif self.sampling.lower() == 'gq':

            M = self.sampling_options.get('M')
            if M is None:
                raise ValueError("The sampling option 'M' must be specified for Gauss quadrature sampling.")

            p_standard, w = self.distribution.polys.tensor_gauss_quadrature(M)
            self.samples = self.map_to_model_space(p_standard)
            self.weights = w

        elif self.sampling.lower() == 'gq-induced':

            K = self.sampling_options.get('K')
            if K is None:
                raise ValueError("The sampling option 'K' must be specified for induced sampling on Gauss quadrature grids.")

            p_standard = self.distribution.opolys.idist_gq_sampling(K, self.indices, M=self.sampling_options.get('M'))

            self.samples = self.map_to_model_space(p_standard)
            self.weights = self.christoffel_weights()

        else:
            raise ValueError("Unsupported sample type '{0}' for input\
                              sample_type".format(self.sampling))

    def integration_weights(self):
        """
        Generates sample weights associated to integration."
        """

        if self.training == 'wlsq':

            p_standard = self.map_to_standard_space(self.samples)

            V = self.distribution.polys.eval(p_standard, self.index_set.get_indices())

            weights = self.christoffel_weights()

            # Should replace with more well-conditioned procedure
            rhs = np.zeros(V.shape[1])
            ind = np.where(np.linalg.norm(self.index_set.get_indices(), axis=1)==0)[0]
            rhs[ind] = 1.
            b = np.linalg.solve((V.T @ np.diag(weights) @ V), rhs)

            return weights * (V @ b)

    def christoffel_weights(self):
        """
        Generates sample weights associated to Christoffel preconditioning.
        """
        p_standard = self.distribution.transform_standard_dist_to_poly.map(
                    self.distribution.transform_to_standard.map(self.samples))

        V = self.distribution.polys.eval(p_standard, self.index_set.get_indices())

        return 1/(np.sum(V**2, axis=1))

    def build_pce_wlsq(self):
        """
        Performs a (weighted) least squares PCE surrogate using saved samples
        and model output.
        """

        p_standard = self.map_to_standard_space(self.samples)

        V = self.distribution.polys.eval(p_standard, self.index_set.get_indices())

        coeffs, residuals = weighted_lsq(V, self.model_output, self.weights)

        self.accuracy_metrics['loocv'] = lstsq_loocv_error(V, self.model_output,
                                                           self.weights)
        self.accuracy_metrics['residuals'] = residuals

        self.coefficients = coeffs
        self.p = self.samples  # Should get rid of this.

        return residuals

    def identify_bulk(self, delta=0.5):
        """
        Performs (adaptive) bulk chasing for refining polynomial spaces.
        Returns the indices associated with a delta-bulk of a OMP-type
        indicator.
        """

        assert 0 < delta <= 1
        indtol = 1e-12

        rmargin = self.index_set.get_reduced_margin()
        indicators = np.zeros(rmargin.shape[0])

        p_standard = self.distribution.transform_standard_dist_to_poly.map(
                    self.distribution.transform_to_standard.map(self.samples))

        # Vandermonde-like matrices for current and margin indices
        V = self.distribution.polys.eval(p_standard, self.index_set.get_indices())
        Vmargin = self.distribution.polys.eval(p_standard, rmargin)

        Vnorms = np.sum(V**2, axis=1)

        residuals = ((V @ self.coefficients) - self.model_output)

        # OMP-style computation of indicator functions
        for m in range(rmargin.shape[0]):
            norms = 1/Vnorms + Vmargin[:, m]**2
            indicators[m] = np.linalg.norm((Vmargin[:, m]*norms).T @ residuals)**2

        if np.sum(indicators) <= indtol:
            print('Current residual error too small: Not adding indices')
            return
        else:
            indicators /= np.sum(indicators)

        # Sort by indicator, and return top indicators that contribute to the
        # fraction delta of unity
        sorted_indices = np.argsort(indicators)[::-1]
        sorted_cumulative_indicators = np.cumsum(indicators[sorted_indices])

        bulk_size = np.argmax(sorted_cumulative_indicators >= delta) + 1

        return rmargin[sorted_indices[:bulk_size], :]

    def augment_samples_idist(self, K, weights=None, fast_sampler=True):
        """
        Augments random samples from induced distribution. Typically done via
        an adaptive refinement procedure. As such some inputs can be given to
        customize how the samples are drawn in the context of adaptivity:

          K: how many samples to add (required)
          weights: a discrete probability distribution on
              self.index_set.get_indices() that describes how the induced
              distrubtion is sampled. Default is uniform.
        """

        return self.distribution.polys.idist_mixture_sampling(K,
                                                              self.index_set.get_indices(),
                                                              weights=weights,
                                                              fast_sampler=fast_sampler)

    def adapt_expressivity(self, max_new_samples=10, **chase_bulk_options):
        """
        Adapts the PCE approximation by increasing expressivity.
        (Intended to combat residual error.)
        """

        from numpy.linalg import norm

        Mold = self.samples.shape[0]
        indices = []
        sample_count = []
        KK = self.accuracy_metrics['residuals'].size
        residuals = [norm(self.accuracy_metrics['residuals'])/np.sqrt(KK), ]
        loocv = [norm(self.accuracy_metrics['loocv'])/np.sqrt(KK), ]
        while self.samples.shape[0] < max_new_samples + Mold:
            samples_left = max_new_samples + Mold - self.samples.shape[0]
            a, b = self.chase_bulk(max_new_samples=samples_left, **chase_bulk_options)
            indices.append(self.index_set.get_indices()[-a:, :])
            sample_count.append(b)
            residuals.append(norm(self.accuracy_metrics['residuals'])/np.sqrt(KK))
            loocv.append(norm(self.accuracy_metrics['loocv'])/np.sqrt(KK))

        return residuals, loocv, indices, sample_count

    def adapt_robustness(self, max_new_samples=10, verbosity=1):
        """
        Adapts the PCE approximation by increasing robustness.
        (Intended to combat cross-validation error.)
        """

        # Just add new samples
        Mold = self.samples.shape[0]
        self.generate_samples(new_samples=max_new_samples)

        # Resample model
        self.model_output = np.vstack((self.model_output,
                      np.zeros([max_new_samples, self.model_output.shape[1]])))
        for ind in range(Mold, Mold+max_new_samples):
            self.model_output[ind, :] = self.model(self.samples[ind, :])

        old_accuracy = self.accuracy_metrics.copy()
        self.build_pce_wlsq()

        KK = np.sqrt(self.model_output.shape[1])
        if verbosity > 0:
            errstr = "new samples: {0:6d}\n  \
                      old residual: {1:1.3e},  old loocv: {2:1.3e}\n  \
                      new residual: {3:1.3e},  new loocv: {4:1.3e}\
                      ".format(max_new_samples,
                               np.linalg.norm(old_accuracy['residuals']/KK),
                               np.linalg.norm(old_accuracy['loocv']/KK),
                               np.linalg.norm(self.accuracy_metrics['residuals']/KK),
                               np.linalg.norm(self.accuracy_metrics['loocv'])/KK)
            print(errstr)

    def chase_bulk(self, delta=0.5, max_new_samples=None, max_new_indices=None,
                   add_rule=None, mult_rule=None, verbosity=1):
        """
        Performs adaptive bulk chasing, which (i) adds the most "important"
        indices to the polynomial index set, (ii) takes more samples, (iii)
        updates the PCE approximation, including statistics and error metrics.

        Args:
            max_new_samples (int): Maximum number of new samples to add.
                Defaults to None.
            max_new_indices (int): Maximum number of new PCE indices to add.
                Defaults to None.
            add_rule (int): Specifies number of samples added as a function
                of number of added indices. Nsamples = Nindices + add_rule.
                Defaults to None.
            mult_rule (float): Specifies number of samples added as a function
                of number of added indices. Nsamples = 
                int(Nindices * add_rule).  Defaults to None.
        """

        if (max_new_samples is not None) and (max_new_indices is not None):
            assert False, "Cannot specify both new sample and new indices max"

        if (add_rule is not None) and (mult_rule is None):
            samplefun = lambda Nindices: int(Nindices + add_rule)
        elif (add_rule is None) and (mult_rule is not None):
            samplefun = lambda Nindices: int(Nindices * mult_rule)
        elif (add_rule is None) and (mult_rule is None):
            samplefun = lambda Nindices: int(Nindices + 2)
        else:
            assert False, 'Cannot specify both an '\
                          'additive and multiplicative rule'

        indices = self.identify_bulk(delta=delta)

        # Determine number of indices we augment by
        if max_new_samples is not None:   # Limited by sample count
            assert max_new_samples > 0
            Nindices = len(indices)

            while samplefun(Nindices) > max_new_samples:
                Nindices -= 1

            # Require at least 1 index to be added.
            Nindices = max(1, Nindices)

        elif max_new_indices is not None:  # Limited by number of indices
            Nindices = max_new_indices

        else:  # No limits: take all indices
            Nindices = len(indices)

        assert Nindices > 0

        L = self.index_set.size()
        weights = np.zeros(L + Nindices)

        # Assign 50% weight to new indices
        weights[:L] = 0.5 / L
        weights[L:] = 0.5 / Nindices

        # Add indices to index set
        self.index_set.augment(indices[:Nindices, :])

        # Add new samples
        Mold = self.samples.shape[0]
        Nsamples = samplefun(Nindices)

        self.weights = weights
        self.generate_samples(new_samples=Nsamples)

        # Resample model
        self.model_output = np.vstack((self.model_output,
                                       np.zeros([Nsamples, self.model_output.shape[1]])))
        for ind in range(Mold, Mold+Nsamples):
            self.model_output[ind, :] = self.model(self.samples[ind, :])

        old_accuracy = self.accuracy_metrics.copy()
        self.build_pce_wlsq()

        KK = np.sqrt(self.model_output.shape[1])
        if verbosity > 0:
            errstr = ('new indices: {0:6d},   new samples: {1:6d}\n'
                      'old residual: {2:1.3e},  old loocv: {3:1.3e}\n'
                      'new residual: {4:1.3e},  new loocv: {5:1.3e}'
                      ).format(Nindices, Nsamples,
                               np.linalg.norm(old_accuracy['residuals']/KK),
                               np.linalg.norm(old_accuracy['loocv']/KK),
                               np.linalg.norm(self.accuracy_metrics['residuals']/KK),
                               np.linalg.norm(self.accuracy_metrics['loocv'])/KK)
            print(errstr)

        return Nindices, Nsamples

    def build(self, model=None, model_output=None, **options):
        """Builds PCE from sampling and approximation settings.

        Args:
            model: A pointer to a function with the syntax xi ---> model(xi),
              which returns a vector corresponding to the model evaluated at
              the stochastic parameter value xi. The input xi to the model
              function should be a vector of size self.dim, and the output
              should be a 1D numpy array. If model_output is None, this is
              required. If model_output is given, this is ignored.
            model_output: A numpy.ndarray corresponding to the output of the
              model at the sample locations specified by self.samples. This is
              required if the input model is None.
        Returns:
            None:
        """

        self.check_distribution()
        self.check_indices()

        # Samples on standard domain
        if self.samples is None:
            self.generate_samples(**options)
        else:
            pass  # User didn't specify samples now, but did previously

        if self.model_output is None:  # We need to generate data
            if model_output is None:

                if model is None:
                    raise ValueError('Must input argument "model" or "model_output".')
                else:
                    self.model = model

                for ind in range(self.samples.shape[0]):
                    if model_output is None:
                        model_output = model(self.samples[ind, :])
                        M = model_output.size
                        model_output = np.concatenate([model_output.reshape([1, M]),
                                                      np.zeros([self.samples.shape[0]-1, M])], axis=0)
                    else:
                        model_output[ind, :] = model(self.samples[ind, :])

            self.model_output = model_output

        else:
            pass  # We'll assume the user did things correctly.

        # For now, we only have 1 method:
        if self.training == 'wlsq':
            return self.build_pce_wlsq()
        else:
            raise ValueError('Unrecongized training directive "{0:s}"'.format(self.training))

    def assert_pce_built(self):
        if self.coefficients is None:
            raise ValueError('First build the PCE with pce.build()')

    def mean(self):
        """Returns PCE mean.

        Returns:
            numpy.ndarray: A vector containing the PCE mean, of size equal to the size
              of the vector of the model output.
        """

        self.assert_pce_built()
        return self.coefficients[0, :]

    def stdev(self):
        """
        Returns PCE standard deviation

        Returns:
            numpy.ndarray: A vector containing the PCE standard deviation, of size
              equal to the size of the vector of the model output.
        """

        self.assert_pce_built()
        return np.sqrt(np.sum(self.coefficients[1:, :]**2, axis=0))

    def pce_eval(self, p, components=None):
        """Evaluates the PCE at particular parameter locations.

        Args:
            p: An array (satisfying p.shape[1]==self.dim) containing a set of
              parameter points at which to evaluate the PCE prediction.
            components: An array of non-negative integers specifying which
              indices in the model output to compute. Other indices are
              ignored. If given as None (default), then all components are
              computed.
        Returns:
            numpy.ndarray: An array containing evaluations (predictions) from the PCE
            emulator. If the input components is None, this array is of size (
            self.p.shape[0] x self.coefficients.shape[1] ). Otherwise, the
            second dimension is of size components.size.
        """

        self.assert_pce_built()
        p_std = self.distribution.transform_standard_dist_to_poly.map(
                    self.distribution.transform_to_standard.map(p))

        if components is None:
            return np.dot(self.distribution.polys.eval(p_std,
                                                       self.index_set.
                                                            get_indices()),
                                                       self.coefficients)
        else:
            return np.dot(self.distribution.polys.eval(p_std,
                                                       self.index_set.
                                                            get_indices()),
                                                       self.coefficients[:, components])

    eval = pce_eval

    def quantile(self, q, M=100):
        """
        Computes q-quantiles using M-point Monte Carlo sampling.
        """

        self.assert_pce_built()
        q = to_numpy_array(q)

        # Maximum number of floats generated at any given time
        MF = max([int(1e6), M, self.distribution.dim])

        # How many model degrees of freedom we can consider at any time
        pce_batch_size = floor(MF/M)

        quantiles = np.zeros([len(q), self.coefficients.shape[1]])

        pce_counter = 0
        p = self.distribution.MC_samples(M)

        while pce_counter < self.coefficients.shape[1]:
            end_ind = min([self.coefficients.shape[1], pce_counter + pce_batch_size])
            inds = range(pce_counter, end_ind)
            ensemble = self.pce_eval(p, components=inds)



            if version_lessthan(np, '1.15'):
                quantiles[:, inds] = mquantiles(ensemble, q, axis=0)
            else:
                quantiles[:, inds] = np.quantile(ensemble, q, axis=0)

            pce_counter = end_ind

        return quantiles

    def total_sensitivity(self, dim_indices=None, vartol=1e-16):
        """
        Computes total sensitivity associated to dimensions dim_indices from
        PCE coefficients. dim_indices should be a list-type containing
        dimension indices.

        The output is len(js) x self.coefficients.shape[1]
        """

        self.assert_pce_built()

        if dim_indices is None:
            dim_indices = range(self.distribution.dim)

        dim_indices = np.asarray(dim_indices, dtype=int)

        indices = self.index_set.get_indices()
        # variance_rows = np.linalg.norm(indices, axis=1) > 0.

        # variances = np.sum(self.coefficients[variance_rows,:]**2, axis=0)
        variance = self.stdev()**2
        total_sensitivities = np.zeros([dim_indices.size, self.coefficients.shape[1]])

        # Return 0 sensitivity if the variance is 0.
        zerovar = variance < vartol

        for (qj, j) in enumerate(dim_indices):

            total_sensitivities[qj, ~zerovar] = np.sum(self.coefficients[np.ix_(indices[:, j] > 0, ~zerovar)]**2,
                                                       axis=0) / variance[~zerovar]

        return total_sensitivities

    def global_sensitivity(self, dim_lists=None, interaction_orders=None, vartol=1e-16):
        """
        Computes global sensitivity associated to dimensional indices dim_lists
        from PCE coefficients.

        dim_lists should be a list of index lists. The global sensitivity for each
        index list is returned.

        interaction_orders (list): Computes sensitivities corresponding to
        variable interactions for the specified orders. E.g., order 2 implies
        binary interactions, 3 is ternary interactions, [2,3] computes both of
        these orders.

        The output is len(dim_lists) x self.coefficients.shape[1]
        """

        d = self.distribution.dim

        return_dim_lists = True
        # If dim_lists is given, ignore interaction_orders
        if dim_lists is not None:
            if interaction_orders is not None:
                print("Ignoring input 'interaction_orders' since 'dim_lists' \
                       was specified.")
            return_dim_lists = False
        else:
            if interaction_orders is None: # Assume all interactions are requested
                interaction_orders = range(1, d+1)
            else:
                try:
                    iter(interaction_orders)
                except: # Assume an int is given
                    interaction_orders = [interaction_orders,]
            dim_lists = list(chain.from_iterable(combinations(range(d), r) for r in interaction_orders))

        # unique_rows = np.vstack({tuple(row) for row in lambdas})
        # # Just making sure
        # assert unique_rows.shape[0] == lambdas.shape[0]

        indices = self.index_set.get_indices()
        # variance_rows = np.linalg.norm(indices, axis=1) > 0.
        # assert np.sum(np.invert(variance_rows)) == 1

        variance = self.stdev()**2
        global_sensitivities = np.zeros([len(dim_lists), self.coefficients.shape[1]])
        dim = self.distribution.dim

        # Return 0 sensitivity if the variance is 0.
        zerovar = variance < vartol

        for (qj, j) in enumerate(dim_lists):
            jc = np.setdiff1d(range(dim), j)
            inds = np.logical_and(np.all(indices[:, j] > 0, axis=1),
                                  np.all(indices[:, jc] == 0, axis=1))

            global_sensitivities[qj, ~zerovar] = np.sum(self.coefficients[np.ix_(inds, ~zerovar)]**2,
                                                        axis=0) / variance[~zerovar]

        if return_dim_lists:
            return global_sensitivities, dim_lists
        else: 
            return global_sensitivities

    def global_derivative_sensitivity(self, dim_list):
        """
        Computes global derivative-based sensitivity indices. For a
        PCE with respect to a :math:`d`-dimensional random variable :math:`Z`,
        then this senstivity index along dimension :math:`i` is defined as

        .. math::

          S_i \\coloneqq E \\left[ p(Z) \\right] = \\int p(z) \\omega(z) d z,

        where :math:`E[\\cdot]` it expectation operator, :math:`p` is the PCE
        emulator, and :math:`\\omega` is the probability density function for
        the random variable :math:`Z`.

        These sensitivity indices measure the average rate-of-change of the PCE
        response with respect to dimension :math:`i`.

        Args:
            dim_lists: A list-type iterable with D entries, containing
              dimensional indices in 0-based indexing. All entries must be
              between 0 and self.distribution.dim.
        Returns:
            S: DxK array, where each row corresponds to the sensitivity index
              :math:`S_i` across all K features of the PCE model.
        """

        indices = self.index_set.get_indices()
        assert all([0 <= dim <= self.distribution.dim-1 for dim in dim_list])

        D = len(dim_list)

        S = np.zeros([D, self.coefficients.shape[1]])

        all_dims = range(self.distribution.dim)

        # TODO: make map compositions default in PCE
        composed_map = self.distribution.transform_standard_dist_to_poly.compose(
                         self.distribution.transform_to_standard)

        # Precompute derivative expansion matrices
        M = self.index_set.max_univariate_degree()
        Cs = [None, ]*self.distribution.dim
        for q in range(self.distribution.dim):
            Cs[q] = self.distribution.\
                         polys.\
                         get_univariate_derivative_expansion(q, 1, M, 0)

        for ind, dim in enumerate(dim_list):
            # Rows of indices whose non-column-dim entries are 0 contribute
            notdim = [val for val in all_dims if val != dim]
            flags = self.index_set.zero_indices(notdim)

            b0 = 1.
            for val in notdim:
                b0 *= self.distribution.polys.\
                                        get_univariate_recurrence(0, val)[0, 1]

            for q in range(self.distribution.dim):
                S[ind, :] += (composed_map.A[q, dim] *
                              Cs[q][indices[flags, dim]].T @
                              self.coefficients[flags, :]).flatten()

            S[ind, :] *= b0

        return S



import numpy as np

from UncertainSCI.families import jacobi_recurrence_values,\
                                  jacobi_weight_normalized

from UncertainSCI.opoly1d import linear_modification, quadratic_modification
from UncertainSCI.opoly1d import gauss_quadrature_driver

# from UncertainSCI.utils.compute_subintervals import compute_subintervals


def compute_subintervals(a, b, singularity_list):
    """
    Returns an M x 4 numpy array, where each row contains the left-hand point,
    right-hand point, left-singularity strength, and right-singularity
    strength.
    """

    # Tolerance for resolving internal versus boundary singularities.
    tol = 1e-12
    singularities = np.array([entry[0] for entry in singularity_list])
    strength_left = np.array([entry[1] for entry in singularity_list])
    strength_right = np.array([entry[2] for entry in singularity_list])

    # We can discard any singularities that lie to the left of a or the right
    # of b
    discard = []
    for (ind, s) in enumerate(singularities):
        if s < a-tol or s > b+tol:
            discard.append(ind)

    singularities = np.delete(singularities, discard)
    strength_left = np.delete(strength_left, discard)
    strength_right = np.delete(strength_right, discard)

    # Sort remaining valid singularities
    order = np.argsort(singularities)
    singularities = singularities[order]
    strength_left = strength_left[order]
    strength_right = strength_right[order]

    # Make sure there aren't doubly-specified singularities
    if np.any(np.diff(singularities) < tol):
        raise ValueError("Overlapping singularities were specified. \
                          Singularities must be unique")

    S = singularities.size

    if S > 0:

        # Extend the singularities lists if we need to add interval endpoints
        a_sing = np.abs(singularities[0] - a) <= tol
        b_sing = np.abs(singularities[-1] - b) <= tol

        # Figure out if singularities match endpoints
        if not b_sing:
            singularities = np.hstack([singularities, b])
            strength_left = np.hstack([strength_left, 0])
            strength_right = np.hstack([strength_right, 0])  # Doesn't matter
        if not a_sing:
            singularities = np.hstack([a, singularities])
            strength_left = np.hstack([0, strength_left])  # Doesn't matter
            strength_right = np.hstack([0, strength_right])  # Doesn't matter

        # Use the singularities lists to identify subintervals
        S = singularities.size
        subintervals = np.zeros([S-1, 4])
        for q in range(S-1):
            subintervals[q, :] = [singularities[q], singularities[q+1],
                                  strength_right[q], strength_left[q+1]]

    else:

        subintervals = np.zeros([1, 4])
        subintervals[0, :] = [a, b, 0, 0]

    return subintervals


def gq_modification(integrand, a, b, N, roots=np.zeros(0),
                    quadroots=np.zeros(0), Nmax=100,
                    gamma=(0, 0), leading_coefficient=1.):
    """
    Uses Gaussian quadrature to approximate

      \\int_a^b q(x) integrand(x) dx,

    where integrand is an input function, and q is a polynomial defined as,

      q(x) = leading_coefficient * prod_{j=1}^R (x - roots[j-1]) *
                                   prod_{q=1}^Q (x - quadroots[j-1])**2,

    where R = roots.size and Q = quadroots.size, and it is assumed that no
    entries of roots lie inside the interval (a,b).  The coefficient
    leading_coefficient can be negative.

    The procedure rewrites the integral using the input gamma as

      \\int_a^b (integrand(x) / w(x)) q(x) w(x) dx,

    where w is a mapped Jacobi weight function:

        w(x) = v(r(x)),    gamma = (alpha, beta)

    where r(x) affinely maps (a,b) to (-1,1), and v(r) is the (alpha,beta)
    Jacobi weight function in UncertainSCI.families.jacobi_weight_normalized.

    The procedure then performs measure modifications on w, absorbing q into
    the measure. An N-point Gaussian quadrature rule with respect to this
    modified measure is used to integrate (integrand / w).
    """

    assert (a < b) and (N > 0)

    # If the gamma parameters are > 0, then we write
    #   gamma[0] = gam + G,
    # where gam \in (-1,0) and G is a positive integer. And similarly for
    # gamma[1].
    #
    # Then we take the Jacobi weight to be w associated with gam,
    # and set N, Nmax += G
    for ind in range(2):
        if gamma[ind] > 0:
            G = np.ceil(gamma[ind])
            gam = gamma[ind] - G
            N += int(G)
            Nmax += int(G)
            gamma[ind] = gam

    assert (gamma[0] <= 0.) and (gamma[1] <= 0.)
    Nmax = max(Nmax, N)

    R = roots.size
    Q = quadroots.size

    # Map everything to [-1,1], and scale by Jacobian afterward
    jac = 2/(b-a)

    def map_to_standard(x):  # Maps [a,b] to [-1,1]
        return jac*(x - a) - 1
    # map_to_standard = lambda x: jac*(x - a) - 1

    def map_to_ab(x):  # Maps [-1,1], to [a,b]
        return (x + 1) / jac + a
    # map_to_ab = lambda x: (x+1)/jac + a

    # Recurrence coefficients for appropriate Jacobi probability measure
    ab = jacobi_recurrence_values(Nmax+R+2*Q, gamma[0], gamma[1])
    ab[0, 1] = 1.

    # The sign of q is determined by how many zeros lie to the right of (a,b)
    sgn = (-1)**(np.sum(roots >= b) % 2) if R > 0 else 1.

    # Modify ab by the zeros of q
    # Simultaneously gradually scale by leading_coefficient and jacobians
    sgn *= np.sign(leading_coefficient)
    if (R+2*Q) > 0:
        # Scale each linear modification by C = leading_coefficient**(1/(R+2Q))
        # There is also a jac factor since (x - x0) = 1/jac * (r-r0)
        # Finally, we need the square root of this because ab[0,1] scales like
        # the square root of an integral.
        C = np.exp(np.log(np.abs(leading_coefficient))/(R+2*Q))/jac
        Csqrt = np.sqrt(C)

        for x0 in quadroots:
            ab = quadratic_modification(ab, map_to_standard(x0))
            ab[0, 1] *= C
        for x0 in roots:
            ab = linear_modification(ab, map_to_standard(x0))
            ab[0, 1] *= Csqrt
    else:
        ab[0, 1] *= np.abs(leading_coefficient)

    x, w = gauss_quadrature_driver(ab, N)

    # Gauss quadrature
    integral = np.sum(w * integrand(map_to_ab(x)) /
                      jacobi_weight_normalized(x, gamma[0], gamma[1]))

    # Jacobian factor for the affine map and sign:
    integral *= sgn/jac

    return integral


def gq_modification_adaptive(integrand, a, b, N, N_step=10, tol=1e-12,
                             **kwargs):
    s = gq_modification(integrand, a, b, N, **kwargs)
    s_new = gq_modification(integrand, a, b, N=N + N_step, **kwargs)

    while np.abs(s - s_new) > tol:
        s = s_new
        N += N_step
        s_new = gq_modification(integrand, a, b, N=N, **kwargs)
    return s_new


def gq_modification_composite(integrand, a, b, N,
                              subintervals=np.zeros([0, 4]), adaptive=True,
                              **kwargs):
    """
    Uses a composite quadrature rule where each subinterval uses
    gq_modification to integrate. The integral is split into
    subintervals.shape[0] integrals.

    subintervals is an M x 4 array, where each row has the form

      [left, right, left_sing, right_sing]

    with each subinterval being [left, right], with the appropriate left- and
    right-hand singularity strengths.

    If subintervals is empty, performs integration over an interval specified
    by

      [a, b, 0, 0]

    Typical keyword arguments that should be input to this function are:

        - quadroots
        - roots
        - leading_coefficient

    See gq_modification for a description of these inputs.
    """
    integral = 0.

    if subintervals.shape[0] == 0:
        subintervals = np.zeros([1, 4])
        subintervals[0, :] = [a, b, 0, 0]

    for q in range(subintervals.shape[0]):
        gamma = [subintervals[q, 3], subintervals[q, 2]]
        if adaptive:
            integral += gq_modification_adaptive(integrand, subintervals[q, 0],
                                                 subintervals[q, 1], N,
                                                 gamma=gamma, **kwargs)
        else:
            integral += gq_modification(integrand, subintervals[q, 0],
                                        subintervals[q, 1], N, gamma=gamma,
                                        **kwargs)

    return integral


def gq_modification_unbounded_composite(integrand, a, b, N, singularity_list,
                                        adaptive=True, tol=1e-12, step=1,
                                        **kwargs):
    """
    Functional as the same as util.quad.gq_modification_composite but with a
    unbounded interval [a,b] I expect this method lying in the util.quad, but
    the issue is:

    1. Subintervals are changed when extending the intervals that we're
    integrating on, thus subintervals cannot be a argument, instead, We use
    singularity_list.

    2. Method compute_subintervals in the composite is required to obtain the
    subintervals for different interval [a,b]

    So We temporarily put this method here, maybe change this later for bettter
    consideration.
    """

    if a == -np.inf and b == np.inf:
        # l = -1.; r = 1.
        le, r = -1., 1.
        subintervals = compute_subintervals(le, r, singularity_list)
        integral = gq_modification_composite(integrand, le, r, N, subintervals,
                                             adaptive, **kwargs)

        integral_new = 1.
        while np.abs(integral_new) > tol:
            r = le
            le = r - step
            subintervals = compute_subintervals(le, r, singularity_list)
            integral_new = gq_modification_composite(integrand, le, r, N,
                                                     subintervals, adaptive,
                                                     **kwargs)
            integral += integral_new

        le, r = -1., 1.
        # l = -1.; r = 1.
        integral_new = 1.
        while np.abs(integral_new) > tol:
            le = r
            r = le + step
            subintervals = compute_subintervals(le, r, singularity_list)
            integral_new = gq_modification_composite(integrand, le, r, N,
                                                     subintervals, adaptive,
                                                     **kwargs)
            integral += integral_new

    elif a == -np.inf:
        r = b
        le = b - step
        subintervals = compute_subintervals(le, r, singularity_list)
        integral = gq_modification_composite(integrand, le, r, N, subintervals,
                                             adaptive, **kwargs)
        integral_new = 1.
        while np.abs(integral_new) > tol:
            r = le
            le = r - step
            subintervals = compute_subintervals(le, r, singularity_list)
            integral_new = gq_modification_composite(integrand, le, r, N,
                                                     subintervals, adaptive,
                                                     **kwargs)
            integral += integral_new

    elif b == np.inf:
        le = a
        r = a + step
        subintervals = compute_subintervals(le, r, singularity_list)
        integral = gq_modification_composite(integrand, le, r, N, subintervals,
                                             adaptive, **kwargs)
        integral_new = 1.
        while np.abs(integral_new) > tol:
            le = r
            r = le + step
            subintervals = compute_subintervals(le, r, singularity_list)
            integral_new = gq_modification_composite(integrand, le, r, N,
                                                     subintervals, adaptive,
                                                     **kwargs)
            integral += integral_new

    return integral
